import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt

from eventdetector_ts.plotter import COLOR_PREDICTED, COLOR_TRUE


def plot_prediction(predicted_op: np.ndarray, filtered_predicted_op: np.ndarray) -> None:
    """
    Plot the original and filtered predicted Op
    Args:
        predicted_op (np.ndarray): Predicted Op
        filtered_predicted_op (np.ndarray): Filtered predicted Op

    Returns:
        None
    """
    sns.set(style="ticks", palette="Set2")
    plt.figure(figsize=(8, 6))  # Set the figure size

    # Plot the true and predicted values using Seaborn
    n = len(predicted_op)
    sns.lineplot(x=np.arange(n), y=predicted_op, color=COLOR_TRUE, label='Predicted Op')
    sns.lineplot(x=np.arange(n), y=filtered_predicted_op, color=COLOR_PREDICTED, label='Filtered Predicted Op')

    # Add labels and title to the plot
    plt.xlabel('partitions')
    plt.ylabel('Op')
    plt.title('Predicted Op')
    # Add legend
    plt.legend()
    # Show
    plt.show()



from typing import Union

import numpy as np


def convolve_with_gaussian_kernel(signal: np.ndarray, sigma: Union[int, float], m: int) -> np.ndarray:
    """
    Convolve a signal with a Gaussian kernel.

    Args:
        signal (np.ndarray): The input signal to convolve.
        sigma (Union[int, float]): The standard deviation of the Gaussian kernel.
        m (int): The radius of the kernel.

    Returns:
        np.ndarray: The convolved signal.

    """

    # Create the Gaussian kernel
    kernel = (1 / (np.sqrt(2 * np.pi) * sigma)) * np.exp(-(np.arange(-m, m + 1) ** 2) / (2 * sigma ** 2))
    kernel /= np.sum(kernel)  # Normalize the kernel

    # Perform the convolution
    convolved_signal = np.convolve(signal, kernel, mode='same')

    return convolved_signal



from datetime import datetime

import matplotlib.dates as mdates
from matplotlib.patches import Rectangle

from eventdetector_ts import TimeUnit
from eventdetector_ts.data.helpers import convert_time_to_datetime, get_timedelta


def event_to_rectangle(event, w_s, time_unit: TimeUnit, color, height=1, style="solid"):
    """
    Function to convert an event to a rectangle object for visualization.
    
    Args:
    event (datetime or other): The event timestamp or object.
    w_s (int): The partition size in time units.
    time_unit (TimeUnit): The time unit of the partition size.
    color (str): The color of the rectangle.
    height (int): The height of the rectangle.
    style (str): The linestyle of the rectangle.

    Returns:
        Rectangle: The rectangle object representing the event.

    """
    time = event
    if not isinstance(event, datetime):
        time = convert_time_to_datetime(event, to_timestamp=False)
    w_s_timedelta = get_timedelta(w_s // 2, time_unit)
    start_time = time - w_s_timedelta
    end_time = time + w_s_timedelta

    start_rect = mdates.date2num(start_time)
    end_rect = mdates.date2num(end_time)

    width_rect = end_rect - start_rect
    rect = Rectangle((start_rect, 0), width_rect, height, edgecolor=color, linestyle=style,
                     facecolor='none', linewidth=1)

    return rect



import json
from datetime import datetime, timedelta
from functools import reduce
from typing import Optional, Union, Tuple, Dict

import numpy as np
import pandas as pd
from dateutil.parser import parser
# noinspection PyUnresolvedReferences
from numpy.lib.stride_tricks import as_strided
from pandas.core.dtypes.common import is_datetime64_any_dtype

from eventdetector_ts import TIME_LABEL, FILL_NAN_ZEROS, FILL_NAN_FFILL, FILL_NAN_BFILL, FILL_NAN_MEDIAN, \
    MIDDLE_EVENT_LABEL, TimeUnit
from eventdetector_ts.data.interval import Interval


def overlapping_partitions(data: np.ndarray, width: int, step: int = 1):
    """
    Splits an input numpy array into a set of overlapping partitions.

    Args:
        data: Input numpy array to be split into overlapping partitions
        width: Width of each overlapping partition
        step: The step size between successive partitions (default=1)

    Returns:
        Numpy array of shape (nb_partitions, width, data.ndim), containing the created overlapping partitions.
    """
    if width > data.shape[0]:
        raise ValueError("Partition size cannot be greater than the size of the input data")
    if step > width:
        raise ValueError("Step size cannot be greater than partition size")

    # Compute the parameters for creating the overlapping partitions
    np_partitions = (data.shape[0] - width) // step + 1
    shape = (np_partitions, width) + data.shape[1:]
    strides = (step * data.strides[0],) + data.strides

    # Use as_strided to create the overlapping partitions
    partitioned_array = as_strided(data, shape=shape, strides=strides)

    return partitioned_array


def convert_dataframe_to_overlapping_partitions(
        dataframe: pd.DataFrame,
        width: int,
        step: int,
        fill_method: Optional[str] = None
) -> np.ndarray:
    """
    Converts a given DataFrame to overlapping partitions.

    Args:
        dataframe: Input DataFrame of features
        width: Width of each overlapping partition
        step: The step size between successive partitions
        fill_method: The method to use for filling NaNs. Supported methods are 'zeros', 'ffill', 'bfill', and 'median'.
            If None, NaNs are left as-is. (default=None)

    Returns:
        Numpy array of shape (np_partitions, width, nb_features), containing the created overlapping partitions.
    """

    dataframe = dataframe.copy()
    dataframe.index = pd.to_datetime(dataframe.index)
    dataframe.loc[:, TIME_LABEL] = dataframe.index.to_pydatetime()

    if fill_method == FILL_NAN_ZEROS:
        dataframe = dataframe.fillna(0)
    elif fill_method == FILL_NAN_FFILL:
        dataframe = dataframe.ffill()
    elif fill_method == FILL_NAN_BFILL:
        dataframe = dataframe.bfill()
    elif fill_method == FILL_NAN_MEDIAN:
        dataframe = dataframe.fillna(dataframe.median())
    elif fill_method is not None:
        raise ValueError(f"Unsupported fill method: {fill_method}")

    sw = overlapping_partitions(dataframe.to_numpy(), width=width, step=step)
    return sw


class InvalidArgumentError(ValueError):
    """Raised when an invalid argument is passed to a function or method."""

    def __init__(self, message):
        """
        Initialize a new InvalidArgumentError with the specified error message.

        Args:
            message (str): The error message to display.
        """
        super().__init__(message)


def convert_time_to_datetime(date: Union[str, pd.Timestamp, float, int], to_timestamp: bool = True) -> \
        Union[float, datetime]:
    """
    Converts a date string, pandas Timestamp, or numeric timestamp to a Python datetime or Unix timestamp.

    Args:
        date: The input date as a string, pandas Timestamp, or numeric timestamp.
        to_timestamp: If True (default), return the date as a Unix timestamp (float), otherwise as a Python datetime.

    Returns:
        The input date as a Unix timestamp or Python datetime object.
    """

    if isinstance(date, pd.Timestamp):
        dt = date.to_pydatetime()
    elif isinstance(date, (float, int)):
        dt = datetime.fromtimestamp(date)
    elif isinstance(date, str):
        dt = parser.parse(date, ignoretz=True)
    else:
        raise ValueError(f"Invalid date format {date}. Supported formats are str, pd.Timestamp, float, and int.")

    if to_timestamp:
        return (dt - datetime(1970, 1, 1)).total_seconds()
    return dt


def num_columns(lst: list) -> int:
    """
    Returns the number of columns in a list.

    Args:
        lst (list): The list to check.

    Returns:
        int: The number of columns in the list.
    """

    if not lst:
        # if the list is empty return 0
        return 0
    elif isinstance(lst[0], list):
        # if the first element of the list is a list, return the length of the first list
        return len(lst[0])
    else:
        # otherwise return 1, because the list has only one column
        return 1


def compute_middle_event(events: Union[list, pd.DataFrame]) -> pd.DataFrame:
    """
    Computes the middle date of events and returns it as a DataFrame.

    Args:
        events (Union[list, pd.DataFrame]): A list or pandas DataFrame containing the starting and ending
            dates of events.

    Returns:
        pd.DataFrame: A pandas DataFrame with a single column containing the middle dates of events.
    """
    column1 = "Starting Date"
    column2 = "Ending Date"
    is2d = True

    if isinstance(events, list):
        nb_columns = num_columns(events)
        if nb_columns == 2:
            df = pd.DataFrame(events, columns=[column1, column2])
        elif nb_columns == 1:
            df = pd.DataFrame(events, columns=[column1])

            is2d = False
        else:
            raise ValueError(
                f"The list of events is not compatible. The number of columns {nb_columns} should not exceed 2.")
    elif isinstance(events, pd.DataFrame):
        df = events
        columns = events.columns
        if len(columns) == 2:
            df = df.rename(columns={columns[0]: column1, columns[1]: column2})
        elif len(columns) == 1:
            is2d = False
            df = df.rename(columns={columns[0]: column1})
        else:
            raise ValueError("The dataframe of events in not compatible, columns should not exceed 2")
    else:
        raise ValueError("The events argument must be a list or pandas DataFrame.")

    df[column1] = pd.to_datetime(df[column1])
    if is2d:
        df[column2] = pd.to_datetime(df[column2])

    if is2d:
        df[column1] = df[column1].apply(lambda x: convert_time_to_datetime(x) / 2)
        df[column2] = df[column2].apply(lambda x: convert_time_to_datetime(x) / 2)
        df[MIDDLE_EVENT_LABEL] = df[column1] + df[column2]
    else:
        df[MIDDLE_EVENT_LABEL] = df[column1].apply(lambda x: convert_time_to_datetime(x))

    df[MIDDLE_EVENT_LABEL] = df[MIDDLE_EVENT_LABEL].apply(lambda x: datetime.utcfromtimestamp(x))
    df = df[[MIDDLE_EVENT_LABEL]]
    df = df.sort_values(by=MIDDLE_EVENT_LABEL)
    return df


def remove_close_events(events_df: pd.DataFrame, delta_unit_time: int, unit: TimeUnit) -> pd.DataFrame:
    """
    Removes events from a DataFrame that occur too close together.

    Args:
        unit: The time unit
        events_df: A pandas DataFrame containing events with a column named 'middle_event'.
        delta_unit_time: A integer representing the minimum time in unit time between events.

    Returns:
        A pandas DataFrame with close events removed.
    """

    # Convert delta to timedelta
    delta = get_timedelta(delta_unit_time, unit)

    # List to hold indices of events to delete
    events_to_delete = []

    # Loop through all events
    for i in range(len(events_df)):
        # Get middle time of the current event
        mid_time = events_df.iloc[i][MIDDLE_EVENT_LABEL]

        # Skip current event if it's already marked for deletion
        if i in events_to_delete:
            continue

        # Loop through all remaining events
        for j in range(i + 1, len(events_df)):
            # Get middle time of the next event
            mid_time1 = events_df.iloc[j][MIDDLE_EVENT_LABEL]

            # If the next event is too close to the current event, mark it for deletion
            if (mid_time1 - mid_time) <= delta:
                events_to_delete.append(j)
            else:
                break

    # Drop events that were marked for deletion
    return events_df.drop(events_df.index[events_to_delete])


def convert_events_to_intervals(events_df: pd.DataFrame, w_s: int, unit: TimeUnit) -> list[Interval]:
    """
    Convert events from a pandas DataFrame to intervals.

    Args:
        events_df (pd.DataFrame): DataFrame containing the events' data.
        w_s (int): The overlapping partition size in unit time.
        unit: The unit time

    Returns:
        list[Interval]: A list of intervals.

    """
    # Create an empty list to store the intervals
    events_intervals = []

    # Loop over the events in the DataFrame
    for i in range(len(events_df)):
        # Get the middle event time
        mid_time = events_df.iloc[i][MIDDLE_EVENT_LABEL]

        # Compute the radius of the interval based on the overlapping partition size
        radius = get_timedelta(delta_unit_time=w_s // 2, unit=unit)

        # Create an interval with the middle event time at the center
        interval = Interval(mid_time - radius, mid_time + radius)

        # Add the interval to the list of intervals
        events_intervals.append(interval)

    # Return the list of intervals
    return events_intervals


def get_union_times_events(events_df: pd.DataFrame, partition_size: int, unit_time: TimeUnit) -> pd.DatetimeIndex:
    """
    Given a DataFrame of events and a time partition size in unit time, computes a DatetimeIndex of all times during 
    which at least one event was taking place.

    Args:
        events_df (pd.DataFrame): A DataFrame containing at least a MIDDLE_EVENT_LABEL column with the datetime
            of each event.
        partition_size (int): The size of the time partition to consider before and after each event.
        unit_time (TimeUnit): The unit time

    Returns:
        pd.DatetimeIndex: A DatetimeIndex of all times during which at least one event was taking place.
    """

    times_during_events = []
    previous_range = None
    for i, event_time in enumerate(events_df[MIDDLE_EVENT_LABEL]):
        start_time = event_time - get_timedelta(partition_size, unit=unit_time)
        end_time = event_time + get_timedelta(partition_size, unit=unit_time)
        # Generate a list of dates between start_time and end_time with a frequency of exactly (end_time - start_time).
        # This ensures that the last date is exactly equal to end_time (useful when we generate overlapping ranges).
        dates_between = pd.date_range(start=start_time, end=end_time, freq=end_time - start_time)

        if previous_range is None:
            times_during_events.append(dates_between)
            previous_range = dates_between
        else:
            # Check if the current range overlaps with the previous one.
            ranges_overlap = max(previous_range[0], previous_range[-1]) < min(dates_between[0], dates_between[-1])
            if not ranges_overlap:
                # If the ranges don't overlap, then we need to merge the previous and current ranges.
                merged_range = pd.date_range(start=previous_range[0], end=dates_between[-1],
                                             freq=dates_between[-1] - previous_range[0])
                # Replace the last range we added to the list with the merged range.
                times_during_events[-1] = merged_range
                previous_range = merged_range
            else:
                previous_range = dates_between
                times_during_events.append(dates_between)

    # Use the reduce function to combine all the overlapping ranges we generated.
    union_ranges = reduce(lambda x, y: x.union(y), times_during_events)
    # Remove any timezone information from the resulting DatetimeIndex, if present.
    union_ranges = union_ranges.tz_localize(None)
    return union_ranges


def get_dataset_within_events_times(data_set: pd.DataFrame, events_times: pd.DatetimeIndex) -> pd.DataFrame:
    """
    Extracts the data from the given dataset that falls within the specified event times.

    Args:
        data_set: A pandas DataFrame containing the data to extract.
        events_times: A pandas DatetimeIndex containing the times of events.

    Returns:
        A pandas DataFrame containing the data within the specified event times.
    """

    dataset_within_events_times = []

    # Iterate through the event times by pairs
    for i in range(0, len(events_times) - 1, 2):
        partition_start_time = events_times[i]
        partition_end_time = events_times[i + 1]

        # Extract the data within the event time
        data_within_event_time = data_set.loc[partition_start_time: partition_end_time]

        dataset_within_events_times.append(data_within_event_time)

    # Concatenate all the data extracted from events times
    return pd.concat(dataset_within_events_times)


def op(dataset_as_overlapping_partitions: np.ndarray, events_as_intervals: list[Interval]) -> \
        tuple[np.ndarray, np.ndarray]:
    """
    Calculates the "op" value for each overlapping partition in the dataset, based on the overlapping parameter 
    between the partition and a set of events.

    Args: dataset_as_overlapping_partitions: A numpy ndarray containing the overlapping partitions for the dataset, 
    where each overlapping partition is a 2D numpy ndarray containing the data points for the partition and their 
    timestamps. events_as_intervals: A list of Interval objects representing the events in the dataset.

    Returns:
        A tuple containing two values:
            - A numpy ndarray containing the overlapping partitions for the dataset, with the timestamp column removed.
            - A numpy ndarray of floating-point values representing the "op" value
                for each overlapping partition in the dataset.
    """

    # The index of the first event that hasn't been checked yet
    starting_event_index = 0

    # List to store the calculated op values for each overlapping partition
    op_values = []

    # Iterate through each overlapping partition in the dataset
    for partition in dataset_as_overlapping_partitions:
        # Get the start and end times of the current overlapping partition
        partition_start_time = partition[0][-1].to_pydatetime()
        partition_end_time = partition[-1][-1].to_pydatetime()

        # Create an Interval object to represent the current overlapping partition
        partition_interval = Interval(partition_start_time, partition_end_time)

        # Initialize the op value for the current overlapping partition to 0
        current_op_value = 0

        # Iterate through each event that hasn't been checked yet
        for event_index in range(starting_event_index, len(events_as_intervals)):
            # Get the Interval object for the current event
            current_event_interval = events_as_intervals[event_index]

            # If the start time of the current partition is greater than or equal to the end time of the current event,
            # we can skip this event since it doesn't overlap with the current partition
            if partition_interval.start_time >= current_event_interval.end_time:
                starting_event_index = event_index + 1
                continue

            # Calculate the overlapping parameter between the current partition and the current event
            overlapping_parameter = partition_interval.overlapping_parameter(current_event_interval)

            # If the overlapping parameter is 0, there is no overlap between the current partition and the current event
            if overlapping_parameter == 0:
                break

            # Update the op value for the current partition if the overlapping parameter is greater than the current op
            # value
            if overlapping_parameter > current_op_value:
                current_op_value = overlapping_parameter

        # Add the op value for the current partition to the list of op values
        op_values.append(current_op_value)

    # Remove the column containing the timestamps from the overlapping partitions
    dataset_as_overlapping_partitions = np.delete(dataset_as_overlapping_partitions, -1, axis=2)

    # Return the updated overlapping partitions and the op values
    return dataset_as_overlapping_partitions, np.array(op_values)


def get_timedelta(delta_unit_time: int, unit: TimeUnit) -> timedelta:
    """
    Returns a timedelta object with the specified delta_unit_time in the specified TimeUnit.

    Args:
        delta_unit_time: The delta unit time value.
        unit: The TimeUnit enum value representing the unit of time.

    Returns:
        A timedelta object with the specified delta_unit_time in the specified TimeUnit.
    """
    if unit == TimeUnit.MICROSECOND:
        return timedelta(microseconds=delta_unit_time)
    elif unit == TimeUnit.MILLISECOND:
        return timedelta(milliseconds=delta_unit_time)
    elif unit == TimeUnit.SECOND:
        return timedelta(seconds=delta_unit_time)
    elif unit == TimeUnit.MINUTE:
        return timedelta(minutes=delta_unit_time)
    elif unit == TimeUnit.HOUR:
        return timedelta(hours=delta_unit_time)
    elif unit == TimeUnit.DAY:
        return timedelta(days=delta_unit_time)
    elif unit == TimeUnit.YEAR:
        return timedelta(days=delta_unit_time * 365)
    else:
        raise ValueError("Invalid TimeUnit value.")


def get_total_units(timedelta_: timedelta, unit: TimeUnit) -> float:
    if unit == TimeUnit.MICROSECOND:
        return timedelta_.total_seconds() * 1e6
    elif unit == TimeUnit.MILLISECOND:
        return timedelta_.total_seconds() * 1e3
    elif unit == TimeUnit.SECOND:
        return timedelta_.total_seconds()
    elif unit == TimeUnit.MINUTE:
        return timedelta_.total_seconds() / 60
    elif unit == TimeUnit.HOUR:
        return timedelta_.total_seconds() / 3600
    elif unit == TimeUnit.DAY:
        return timedelta_.total_seconds() / (3600 * 24)
    elif unit == TimeUnit.YEAR:
        return timedelta_.total_seconds() / (3600 * 24 * 365.25)
    else:
        raise ValueError("Invalid TimeUnit value.")


def check_time_unit(diff: timedelta) -> Tuple[int, TimeUnit]:
    """
    Method to determine the unit of time of the dataset.

    Args:
        diff (timedelta): The time difference to be checked.

    Returns:
        Tuple[int, TimeUnit]: A tuple with the time value and its unit.
    """

    if diff.total_seconds() >= 31536000:  # 1 year in seconds
        years = int(diff.total_seconds() / 31536000)
        t_s = years
        time_unit = TimeUnit.YEAR
    elif diff.total_seconds() >= 86400:  # 1 day in seconds
        days = int(diff.total_seconds() / 86400)
        t_s = days
        time_unit = TimeUnit.DAY
    elif diff.total_seconds() >= 3600:  # 1 hour in seconds
        hours = int(diff.total_seconds() / 3600)
        t_s = hours
        time_unit = TimeUnit.HOUR
    elif diff.total_seconds() >= 60:  # 1 minute in seconds
        minutes = int(diff.total_seconds() / 60)
        t_s = minutes
        time_unit = TimeUnit.MINUTE
    elif diff.total_seconds() >= 1:
        t_s = int(diff.total_seconds())
        time_unit = TimeUnit.SECOND
    elif diff.total_seconds() * 1000 >= 1:
        t_s = int(diff.total_seconds() * 1000)
        time_unit = TimeUnit.MILLISECOND
    elif diff.total_seconds() * 1000000 >= 1:
        t_s = int(diff.total_seconds() * 1000000)
        time_unit = TimeUnit.MICROSECOND
    else:
        raise ValueError("Could not determine the unit of time of the dataset")

    return t_s, time_unit


def save_dict_to_json(path: str, data: Dict):
    """
    Save a dictionary into a json file
    Args:
        path (str): the path where to store the json file
        data (Dict): the dictionary

    Returns:

    """
    with open(path, 'w') as f:
        json.dump(data, f)


def convert_dataset_index_to_datetime(dataset: pd.DataFrame) -> None:
    """
    Check if the index of the DataFrame dataset is already in the datetime format. If the index is not in datetime 
    format, dataset.index = pd.to_datetime(dataset.index) statement is executed to convert it. 
    
    Args: 
        dataset (pd.DataFrame): A dataset as pandas DataFrame

    Returns:
        None
    """
    if not is_datetime64_any_dtype(dataset.index):
        dataset.index = pd.to_datetime(dataset.index)



# Set up logging
import os
import pprint
import shutil
from typing import Union, Dict

import numpy as np
import pandas as pd
import tensorflow as tf

from eventdetector_ts import FFN, FILL_NAN_ZEROS, TYPE_TRAINING_AVERAGE, STANDARD_SCALER, \
    config_dict, CONFIG_FILE
from eventdetector_ts.data.helpers import compute_middle_event, remove_close_events, \
    convert_events_to_intervals, get_union_times_events, get_dataset_within_events_times, \
    convert_dataframe_to_overlapping_partitions, op, check_time_unit, save_dict_to_json, \
    convert_dataset_index_to_datetime
from eventdetector_ts.metamodel import logger_meta_model
from eventdetector_ts.metamodel.utils import DataSplitter, validate_args
from eventdetector_ts.models.models_builder import ModelCreator
from eventdetector_ts.models.models_trainer import ModelTrainer
from eventdetector_ts.optimization.event_extraction_pipeline import OptimizationData, EventOptimization
from eventdetector_ts.plotter.plotter import Plotter


class MetaModel:
    def __init__(
            self,
            output_dir: str,
            dataset: pd.DataFrame,
            events: Union[list, pd.DataFrame],
            width: int,
            step: int = 1,
            **kwargs
    ):
        """
        Initializes a new instance of the MetaModel class.

        Args:
            output_dir (str): The name or path of the directory where all outputs will be saved.
            If output_dir is a
                folder name, it will create the full path in the current directory.
            dataset (pd.DataFrame): The input dataset as pd.DataFrame.
            events (Union[list, pd.DataFrame]): The input events.
            width (int): The width to be used for creating overlapping partitions.
            step (int): The step size between two successive partitions.
            kwargs (Dict): Optional keyword arguments:
                - t_max (float): The maximum total time related to sigma. The default value is (3 * self.w_s) / 2).
                - delta (int): The maximum time tolerance used to determine the correspondence between a predicted event
                    and its actual counterpart in the true events.
                    The default value is w_s.
                - s_h (float): A step parameter for the peak height threshold h. The default value is 0.05.
                - epsilon (float): A small constant used to control the size of set which contains the top models
                    with the lowest MSE values. The default value is 0.0002.
                - pa (int): The patience for the early stopping algorithm. The default value is 5.
                - t_r (float): The ratio threshold for the early stopping algorithm.
                    The default value is 0.97.
                - time_window Optional[int] = None: The 'time_window' parameter is crucial for controlling the amount 
                    of data used in the dataset.
                    It should be specified as a number of units of time.
                    By default, it is set to None, which means that all available data will be used.
                    However, if a value is provided, the dataset will only include a specific interval of data 
                    around each reference event.
                    This interval consists of data from both the left and right sides of 
                    each event, with a duration equal to the specified 'time_window'.
                    Setting a time_window can offer 
                    several advantages, including speeding up the training process and improving the 
                    neural networks' understanding for rare events.
                - models (List[Union[str, Tuple[str, int]]]): Determines the type of deep learning models to use.
                    If a tuple is passed, it specifies both the model type and the number of instances to run.
                    The default value is [(model, 2) for model in [FFN]].
                - hyperparams_ffn (Tuple[int, int, int]): Specify for the FFN the maximum number of layers,
                    the minimum and the maximum number of neurons per layer.
                    The default value is (3, 64, 256).
                - hyperparams_cnn (Tuple[int, int, int, int, int]): Specify for the CNN the minimum, maximum number
                    of filters, the minimum, the maximum kernel size, and maximum number of pooling layers.
                    The default value is (16, 64, 3, 8 , 2).
                - hyperparams_rnn (Tuple[int, int, int]): Specify for the RNN the maximum number of RNN layers
                    the minimum and the maximum number of hidden units.
                    The default value is (1, 16, 128).
                - hyperparams_mm_network (Tuple[int, int]): Specify for the MetaModel network the number
                    of layers and the number of neurons per layer.
                    The default value is (1, 32).
                - epochs (int): The number of epochs to train different models. The default value is False 256.
                - batch_size (int): The number of samples per gradient update.
                    The default value is 32.
                - fill_nan (str): Specifies the method to use for filling NaN values in the dataset.
                    Supported methods are 'zeros', 'ffill', 'bfill', and 'median'.
                    The default is 'zeros'.
                - type_training (str):Specifies the type of training technique to use for the MetaModel.
                    Supported techniques are 'average' and 'ffn'.
                    The default is 'average'.
                - scaler (str): The type of scaler to use for preprocessing the data.
                    Possible values are "MinMaxScaler", "StandardScaler", and "RobustScaler".
                    Default is "StandardScaler"
                - use_kfold (bool): Whether to use k-fold cross-validation technique or not.
                The default value is False.
                - test_size (float): The proportion of the dataset to include in the test split.
                    Should be a value between 0 and 1. Default is 0.2.
                - val_size (float): The proportion of the training set to use for validation.
                    Should be a value between 0 and 1. Default is 0.2.
                - use_multiprocessing (bool): Whether to use multiprocessing or not for the event exctraction
                    optimization. The default value is False.
                - save_models_as_dot_format (bool): Whether to save the models as a dot format file.
                    The default value is False. If set to True, then you should have graphviz software
                    to be installed on your machine.
        """
        self.step = step
        self.width = width
        self.events = events
        self.dataset = dataset
        self.output_dir = output_dir
        self.kwargs: Dict = kwargs
        self.y = np.empty(shape=(0,))
        self.x = np.empty(shape=(0,))
        self.__compute_and_set_time_sampling()
        self.__set_defaults()
        validate_args(self)
        if self.save_models_as_dot_format:
            logger_meta_model.warning("save_models_as_dot_format is set to true, "
                                      "you should have graphviz software to be installed on your machine.")
        self.__create_output_dir()
        # Create a `ModelCreator` object with the provided models and hyperparameters
        self.model_creator: ModelCreator = ModelCreator(models=self.models, hyperparams_ffn=self.hyperparams_ffn,
                                                        hyperparams_cnn=self.hyperparams_cnn,
                                                        hyperparams_rnn=self.hyperparams_rnn,
                                                        save_models_as_dot_format=self.save_models_as_dot_format,
                                                        root_dir=self.output_dir)
        # Create a `DataSplitter` object with the provided test_size and scaler_type
        self.data_splitter: DataSplitter = DataSplitter(test_size=self.test_size, scaler_type=self.scaler)
        # Create a `ModelTrainer` object with the provided data_splitter, epochs,
        #   batch_size, pa, t_r, use_kfold, val_size, epsilon and save_models_as_dot_format.
        self.model_trainer: ModelTrainer = ModelTrainer(data_splitter=self.data_splitter, epochs=self.epochs,
                                                        batch_size=self.batch_size, pa=self.pa, t_r=self.t_r,
                                                        use_kfold=self.use_kfold,
                                                        val_size=self.val_size, epsilon=self.epsilon,
                                                        save_models_as_dot_format=self.save_models_as_dot_format)
        # class represents the data used for the event extraction pipeline.
        self.optimization_data: OptimizationData = OptimizationData(t_max=self.t_max, w_s=self.w_s, s_s=self.s_s,
                                                                    s_h=self.s_h, delta=self.delta,
                                                                    output_dir=self.output_dir,
                                                                    time_unit=self.time_unit)
        # The Plotter class is responsible for generating and saving plots.
        self.plotter: Plotter = Plotter(root_dir=self.output_dir, time_unit=self.time_unit, w_s=self.w_s)

    def __create_output_dir(self) -> None:
        """
           Check if output_dir is already a complete path, if output_dir is a folder name,
            create the full path in the current directory.

           Returns:
               None
           """

        # Check if output_dir is already a complete path
        if os.path.isabs(self.output_dir):
            if not os.path.exists(self.output_dir):
                logger_meta_model.critical(f"{self.output_dir} does not exists")
                raise ValueError(f"{self.output_dir} does not exists")

        # If output_dir is a folder name, create the full path in the current directory
        else:
            # Get the absolute path of the current directory
            current_directory = os.path.abspath(".")
            self.output_dir = os.path.join(current_directory, self.output_dir)
            if os.path.exists(self.output_dir):
                logger_meta_model.warning(f"The working directory '{self.output_dir}' exists and it will be deleted")
                shutil.rmtree(self.output_dir)
            logger_meta_model.info(f"Creating the working directory at: '{self.output_dir}'")
            os.makedirs(self.output_dir)

        config_dict['output_dir'] = self.output_dir

    def __set_defaults(self) -> None:
        """
        Sets default values for any missing keyword arguments in self.kwargs.

        Returns:
            None
        """
        self.t_max = self.kwargs.get('t_max', (3.0 * self.w_s) / 2)  # the minimum should be equal to w_s
        self.delta = self.kwargs.get('delta', self.w_s)
        self.s_h = self.kwargs.get('s_h', 0.05)
        self.epsilon = self.kwargs.get('epsilon', 0.0002)
        self.pa = self.kwargs.get('pa', 5)
        self.t_r = self.kwargs.get('t_r', 0.97)
        self.time_window = self.kwargs.get('time_window', None)
        self.models = self.kwargs.get('models', [(model, 2) for model in [FFN]])
        for i, model in enumerate(self.models):
            if isinstance(model, str):
                self.models[i] = (model, 2)
            elif isinstance(model, tuple) and len(model) == 1:
                self.models[i] = (model[0], 3)

        self.hyperparams_ffn = self.kwargs.get('hyperparams_ffn', (3, 64, 256))
        self.hyperparams_cnn = self.kwargs.get('hyperparams_cnn', (16, 64, 3, 8, 2))
        self.hyperparams_rnn = self.kwargs.get('hyperparams_rnn', (1, 16, 128))
        self.hyperparams_mm_network = self.kwargs.get('hyperparams_mm_network', (1, 32))
        self.epochs = self.kwargs.get('epochs', 256)
        self.batch_size = self.kwargs.get('batch_size', 32)
        self.fill_nan = self.kwargs.get('fill_nan', FILL_NAN_ZEROS)
        self.type_training = self.kwargs.get('type_training', TYPE_TRAINING_AVERAGE)
        self.scaler = self.kwargs.get('scaler', STANDARD_SCALER)
        self.use_kfold = self.kwargs.get('use_kfold', False)
        self.test_size = self.kwargs.get('test_size', 0.2)
        self.val_size = self.kwargs.get('val_size', 0.2)
        self.use_multiprocessing = self.kwargs.get('use_multiprocessing', False)
        self.save_models_as_dot_format = self.kwargs.get('save_models_as_dot_format', False)

        log_dict = {
            't_max': self.t_max,
            'delta': self.delta,
            's_h': self.s_h,
            'epsilon': self.epsilon,
            'pa': self.pa,
            't_r': self.t_r,
            'time_window': self.time_window,
            'models': self.models,
            'hyperparams_ffn': self.hyperparams_ffn,
            'hyperparams_cnn': self.hyperparams_cnn,
            'hyperparams_rnn': self.hyperparams_rnn,
            'hyperparams_mm_network': self.hyperparams_mm_network,
            'epochs': self.epochs,
            'batch_size': self.batch_size,
            'fill_nan': self.fill_nan,
            'type_training': self.type_training,
            'scaler': self.scaler,
            'use_kfold': self.use_kfold,
            'test_size': self.test_size,
            'val_size': self.val_size,
            'use_multiprocessing': self.use_multiprocessing,
            'save_models_as_dot_format': self.save_models_as_dot_format
        }

        log_message = pprint.pformat(log_dict, indent=4)
        logger_meta_model.warning(log_message)

        config_dict.update({'width': self.width, 'step': self.step, 'batch_size': self.batch_size,
                            'type_training': self.type_training, 'fill_nan': self.fill_nan})

    def __compute_and_set_time_sampling(self) -> None:
        """
        Compute the time sampling of the dataset by calculating the time difference between the first two index values.
                Then set the corresponding parameters: t_s, w_s, and s_s.

        Returns:
            None

        Raises:
            TypeError: If the index of the dataset is not in datetime format.
        """
        try:
            logger_meta_model.info("checks if the index of the dataset is already in the datetime format.")
            convert_dataset_index_to_datetime(self.dataset)
            # Get the first two index values of the dataset
            a = self.dataset.index[0]
            b = self.dataset.index[1]
            # Calculate the time difference between the first two index values
            diff = b - a
            # Check the units of the time difference
            logger_meta_model.info("Computing the time sampling and time unit of the dataset")
            self.t_s, self.time_unit = check_time_unit(diff=diff)
            logger_meta_model.warning(f"The time sampling t_s is {self.t_s} {self.time_unit}s")
            self.w_s = self.t_s * self.width
            self.s_s = self.t_s * self.step
            config_dict['time_unit'] = self.time_unit.__str__()
            config_dict['w_s'] = self.w_s
        except AttributeError:
            logger_meta_model.critical("The dataset is not compatible with the datetime format")
            raise TypeError("The index should be in datetime format.")

    def prepare_data_and_computing_op(self) -> None:
        """
        Prepare the events and dataset for computing op.
        This method will compute the middle event of the given events, remove any close events based on the self.w_s,
            and convert the remaining events to intervals. If a time partition is specified, it will get the union of
            event times and extract the corresponding portion of the dataset.

        The dataset will then be converted to overlapping partitions using the specified width and step size, 
        and the $op$ (overlapping parameter) values will be computed for each partition based on the given intervals.

        Finally, the learning data (overlapping partitions and corresponding $op$ values) will be stored in
            the instance variables x and y.

        Returns:
             None
        """

        logger_meta_model.info("Computes the middle date of events...")
        self.events = compute_middle_event(self.events)

        logger_meta_model.info("Removes events that occur too close together...")
        temp: int = len(self.events)
        self.events = remove_close_events(self.events, self.w_s, self.time_unit)

        logger_meta_model.warning(f"A total of {temp - len(self.events)}/{temp} events were removed due to overlapping")
        logger_meta_model.info("Convert events to intervals...")
        intervals = convert_events_to_intervals(self.events, self.w_s, self.time_unit)

        if self.time_window is not None:
            logger_meta_model.warning(f"time_window is provided = {self.time_window} {self.time_unit}s")
            events_times = get_union_times_events(self.events, self.time_window, self.time_unit)
            self.dataset = get_dataset_within_events_times(self.dataset, events_times)

        logger_meta_model.info("Computing overlapping partitions...")
        overlapping_partitions = convert_dataframe_to_overlapping_partitions(self.dataset, width=self.width,
                                                                             step=self.step,
                                                                             fill_method=self.fill_nan)

        logger_meta_model.info("Computing op...")
        self.x, self.y = op(dataset_as_overlapping_partitions=overlapping_partitions, events_as_intervals=intervals)

        # Convert x and y arrays to float32 for consistency
        self.x = np.asarray(self.x).astype('float32')
        self.y = np.asarray(self.y).astype('float32')

        self.optimization_data.set_overlapping_partitions(overlapping_partitions)
        self.optimization_data.set_true_events(self.events)

    def build_stacking_learning(self) -> None:
        """
        Builds a stacking learning pipeline using the provided models and hyperparameters.

        Returns:
            None
        """

        # Get the number of time steps and features from the x data
        n_time_steps, n_features = self.x.shape[1], self.x.shape[2]
        config_dict['n_time_steps'] = n_time_steps
        inputs = tf.keras.Input(shape=(n_time_steps, n_features), name="input")
        # Call the `create_models` method to create the models
        logger_meta_model.info(f"Create the following models: {list(map(lambda x: x[0], self.models))}")
        self.model_creator.create_models(inputs=inputs)
        logger_meta_model.info("Split the data into training, validation, and test sets and apply "
                               "the specified scaler to each time step...")
        self.data_splitter.split_data_and_apply_scaler(x=self.x, y=self.y)
        logger_meta_model.info("Saves the scalers to disk...")
        self.data_splitter.save_scalers(output_dir=self.output_dir)
        logger_meta_model.info("Fits the created models to the training data...")
        self.model_trainer.fitting_models(self.model_creator.created_models)
        logger_meta_model.info("Saving the best models...")
        self.model_trainer.save_best_models(output_dir=self.output_dir)
        predicted_y, loss, test_y = self.model_trainer.train_meta_model(type_training=self.type_training,
                                                                        hyperparams_mm_network
                                                                        =self.hyperparams_mm_network,
                                                                        output_dir=self.output_dir)
        self.optimization_data.set_predicted_op(predicted_op=predicted_y)
        logger_meta_model.info(f"The loss of the MetaModel is {loss:.4f}")
        self.plotter.set_data_op(test_y=test_y, predicted_y=predicted_y)
        self.plotter.set_losses(train_losses=self.model_trainer.train_losses,
                                val_losses=self.model_trainer.val_losses, train_loss_meta_model=
                                self.model_trainer.train_loss_meta_model,
                                val_loss_meta_model=self.model_trainer.val_loss_meta_model)

    def event_extraction_optimization(self) -> None:
        """
        Run the Event Extraction Optimization process.

        Returns:
            None
        """
        event_optimization: EventOptimization = EventOptimization(optimization_data=self.optimization_data)
        predicted_events, delta_t = event_optimization.max_f1score(use_multiprocessing=self.use_multiprocessing)
        path = os.path.join(self.output_dir, CONFIG_FILE)
        logger_meta_model.info(f"Saving config file into {path}")
        save_dict_to_json(path=path, data=config_dict)
        self.plotter.set_data_events(predicted_events=predicted_events, true_events=self.optimization_data.true_events)
        self.plotter.set_delta_t(delta_t=delta_t)

    def plot_save(self, show_plots: bool = True) -> None:
        """
        Plot the results: losses, true/predicted op, true/predicted events, deltat_t.

        Args:
            show_plots (bool): whether to show the plots or not.
            
        Returns:
            None
        """
        self.plotter.set_show(show=show_plots)
        self.plotter.plot_losses()
        self.plotter.plot_prediction()
        self.plotter.plot_predicted_events()
        self.plotter.plot_delta_t(bins=10)



import logging

import numpy as np
import tensorflow as tf
from keras.utils.io_utils import print_msg
from sklearn.model_selection import KFold


class CustomEarlyStopping(tf.keras.callbacks.Callback):
    """
    Create a custom early stopping callback that stops training when the ratio of current training loss to current
    validation loss is less than a specified ratio for a number of consecutive epochs.

    Args:
        ratio (float): Ratio to compare current train loss and current val loss against.
        patience (int): Number of epochs to wait before stopping training.
        verbose (int, optional): Verbosity level.

    Attributes:
        stopped_epoch (int or None): Last epoch index where training was stopped.
        best (float or None): Best validation loss observed so far.
        best_epoch (int or None): Index of the epoch where the best validation loss was observed.
        ratio (float): Ratio to compare current train loss and current val loss against.
        patience (int): Number of epochs to wait before stopping training.
        verbose (int): Verbosity level.
        wait (int): Number of epochs since the last time the ratio was greater than self.ratio.
        monitor_op (function): Comparison operator for the ratio.
        best_weights (np.ndarray or None): Model weights at the epoch with the best validation loss.
    """

    def __init__(self, ratio: float, patience: int, verbose: int = 1):
        super().__init__()
        self.stopped_epoch = None
        self.best = None
        self.best_epoch = None
        self.ratio = ratio
        self.patience = patience
        self.verbose = verbose
        self.wait = 0
        self.monitor_op = np.greater
        self.best_weights = None

    def on_train_begin(self, logs=None):
        """
        Initialize instance attributes.
        """
        self.wait = 0
        self.best_weights = None
        self.stopped_epoch = 0
        self.best_epoch = 0
        self.best = np.Inf

    def on_epoch_end(self, epoch, logs=None):
        """
        Update the best validation loss and check whether to stop training.
        """
        if logs is not None:
            if self.best_weights is None:
                self.best_weights = self.model.get_weights()

            current_val = logs.get('val_loss')  # Current validation loss
            current_train = logs.get('loss')  # Current training loss
            if current_val is None:
                logging.warning(
                    "Early stopping conditioned on metric `%s` "
                    "which is not available. Available metrics are: %s",
                    'val_loss',
                    ",".join(list(logs.keys())),
                )

            # Update the best validation loss and weights
            if self.monitor_op(self.best, current_val):
                self.best = current_val
                self.best_weights = self.model.get_weights()
                self.best_epoch = epoch

            # If the ratio of current training loss to current validation loss is greater than the specified ratio.
            if self.monitor_op(np.divide(current_train, current_val), self.ratio):
                self.wait = 0
            else:
                # Only check after the first epoch.
                if self.wait >= self.patience and epoch > 0:
                    self.stopped_epoch = epoch
                    self.model.stop_training = True
                    if self.verbose > 0:
                        print_msg(
                            "Restoring model weights from "
                            "the end of the best epoch: "
                            f"{self.best_epoch + 1}."
                        )
                    self.model.set_weights(self.best_weights)
                self.wait += 1

    def on_train_end(self, logs=None):
        """
        Print a message indicating that training was stopped early.
        """
        if logs is not None:
            if self.stopped_epoch > 0 and self.verbose > 0:
                print_msg(
                    f"Epoch {self.stopped_epoch + 1}: early stopping. "
                    "Restoring model weights from "
                    "the end of the best epoch: "
                    f"{self.best_epoch + 1}. "
                    "Best validation loss: "
                    f"{self.best}."
                )


class SelfAttention(tf.keras.layers.Layer):
    """
    Self-Attention layer for Neural Networks
    """

    def __init__(self, units: int, **kwargs) -> None:
        super().__init__()
        self.last_attention_weights = None
        # Instantiate a multi-head attention layer with key dimensionality of units
        # and a single head
        self.mha = tf.keras.layers.MultiHeadAttention(key_dim=units, num_heads=1, **kwargs)
        # Instantiate a normalization layer
        self.layer_norm = tf.keras.layers.LayerNormalization()
        # Instantiate an addition layer
        self.add = tf.keras.layers.Add()

    def call(self, query: tf.Tensor) -> tf.Tensor:
        """
        Apply a self-attention mechanism on the input query and return the output.

        Args:
            query: input tensor to the layer.

        Return:
            output tensor of the layer.
        """
        # Apply multi-head attention on a query
        attn_output, attn_scores = self.mha(
            query=query,
            key=query,
            value=query,
            return_attention_scores=True)

        # Store the attention scores in last_attention_weights for inspection
        self.last_attention_weights = attn_scores

        # Add the attention output to the query and normalize it
        x = self.add([query, attn_output])
        x = self.layer_norm(x)

        return x


def custom_cross_val_score(model: tf.keras.Model, x: np.ndarray, y: np.ndarray, cv: KFold, epochs: int, batch_size: int,
                           callbacks: list) -> np.ndarray:
    """
    A function to perform custom cross-validation for a Keras model.
    
    Args:
        model: A Keras model.
        x: The input data.
        y: The target data.
        cv: A KFold cross-validation object.
        epochs: The number of epochs for training.
        batch_size: The batch size for training.
        callbacks: A list of Keras callbacks.

    Returns:
        The mean of the validation loss across all folds.
    """
    scores = []
    for train_index, val_index in cv.split(x):
        train_x, train_y = x[train_index], y[train_index]
        val_x, val_y = x[val_index], y[val_index]
        history = model.fit(train_x, train_y, epochs=epochs, batch_size=batch_size, callbacks=callbacks,
                            validation_data=(val_x, val_y), verbose=0)
        scores.append(np.min(history.history['val_loss']))
    return np.mean(scores)



import os
import re
from typing import Dict

import joblib
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, MinMaxScaler, RobustScaler

from eventdetector_ts import MIN_MAX_SCALER, ROBUST_SCALER, SCALERS_DIR, FILL_NAN_ZEROS, FILL_NAN_FFILL, \
    FILL_NAN_BFILL, FILL_NAN_MEDIAN, RNN_BIDIRECTIONAL, CONV_LSTM1D, RNN_ENCODER_DECODER, FFN, CNN_RNN, \
    GRU, CNN, SELF_ATTENTION, LSTM, TYPE_TRAINING_AVERAGE, TYPE_TRAINING_FFN, STANDARD_SCALER
from eventdetector_ts.data.helpers import InvalidArgumentError


class DataSplitter:
    """
    A class for splitting and scaling data into training, test sets and applying scalers to each
    time step in the data.
    """

    def __init__(self, test_size: float, scaler_type: str):
        """
        Initialize the DataSplitter object.

        Args:
            test_size: The fraction of data to use for testing.
            scaler_type: The type of scaler to use.
        """

        self.train_x: np.ndarray = np.empty(shape=(0,))
        self.test_x: np.ndarray = np.empty(shape=(0,))
        self.train_y: np.ndarray = np.empty(shape=(0,))
        self.test_y: np.ndarray = np.empty(shape=(0,))
        self.scalers: Dict[int, StandardScaler | MinMaxScaler | ROBUST_SCALER] = {}
        self.test_size: float = test_size
        self.scaler_type: str = scaler_type

    def split_data_and_apply_scaler(self, x: np.ndarray, y: np.ndarray) -> None:
        """
        Split the data into training, validation, and test sets and apply the specified scaler to each time step.

        Args:
            x: The input data with shape (n_samples, n_time_steps, n_features).
            y: The target data with shape (n_samples,).

        Returns:
            A tuple containing the training, validation, and test sets as numpy arrays and a dictionary of scalers.
        """
        assert x.ndim == 3, "x must be a 3D array."
        assert y.ndim == 1, "y must be a 1D array."
        assert x.shape[0] == y.shape[0], "x and y must have the same number of samples."

        # Split the data into training and test sets
        self.train_x, self.test_x, self.train_y, self.test_y = train_test_split(x, y, test_size=self.test_size,
                                                                                shuffle=False)

        n_time_steps = x.shape[1]

        self.scalers = {}
        # Apply scaler to each time step
        for i in range(n_time_steps):
            scaler = StandardScaler()
            if self.scaler_type == MIN_MAX_SCALER:
                scaler = MinMaxScaler()
            elif self.scaler_type == ROBUST_SCALER:
                scaler = RobustScaler()
            self.scalers[i] = scaler
            self.train_x[:, i, :] = self.scalers[i].fit_transform(self.train_x[:, i, :])
            self.test_x[:, i, :] = self.scalers[i].transform(self.test_x[:, i, :])

    def save_scalers(self, output_dir: str) -> None:
        """
        Saves the scalers to disk.

        Args:
            output_dir: the directory where the scalers should be saved

        Returns:
            None
        """
        # Create the directory if it doesn't exist
        scalers_dir = os.path.join(output_dir, SCALERS_DIR)
        if not os.path.exists(scalers_dir):
            os.makedirs(scalers_dir)

        # Save each scaler to disk
        n_time_steps: int = self.test_x.shape[1]
        for i in range(n_time_steps):
            # Generate the path to save the scaler to
            scaler_i_path = os.path.join(scalers_dir, f'scaler_{i}.joblib')
            # Print progress
            print("\rSaving scaling...{}/{}".format(i + 1, n_time_steps), end="")
            # Save the scaler to disk
            joblib.dump(self.scalers[i], scaler_i_path)
        print()


def validate_args(meta_model) -> None:
    """
    Validate the arguments of the MetaModel.

    Args:
        meta_model (MetaModel): A MetaModel instance.

    Returns:
        None

    Raises:
        ValueError: If any of the arguments are invalid.
    """

    if not isinstance(meta_model.width, int) or meta_model.step <= 0:
        raise InvalidArgumentError("step should be a positive integer.")

    if not isinstance(meta_model.width, int) or meta_model.width <= meta_model.step:
        raise InvalidArgumentError(f"width should be greater than {meta_model.step}.")

    if meta_model.dataset is None or meta_model.dataset.empty:
        raise InvalidArgumentError("dataset cannot be None or empty.")
    elif not isinstance(meta_model.dataset, pd.DataFrame):
        raise InvalidArgumentError("dataset should be a pandas DataFrame.")

    if len(meta_model.dataset) < meta_model.width:
        raise InvalidArgumentError("Dataset length is smaller than the given partition width.")

    if meta_model.events is None or (isinstance(meta_model.events, pd.DataFrame) and meta_model.events.empty) or \
            (isinstance(meta_model.events, list) and len(meta_model.events) == 0):
        raise InvalidArgumentError("events is empty or None.")
    elif not isinstance(meta_model.events, (list, pd.DataFrame)):
        raise InvalidArgumentError("events should be a list or a pandas DataFrame.")

    if not re.match("^[a-zA-Z0-9_]+$", meta_model.output_dir):
        raise InvalidArgumentError(
            "Output directory name can only contain alphanumeric characters and underscores.")

    if meta_model.fill_nan not in [FILL_NAN_ZEROS, FILL_NAN_FFILL, FILL_NAN_BFILL, FILL_NAN_MEDIAN]:
        raise InvalidArgumentError(
            f"Invalid method for filling NaN values. Supported methods are"
            f"  {FILL_NAN_ZEROS}, {FILL_NAN_FFILL}, {FILL_NAN_BFILL}, and {FILL_NAN_MEDIAN}.")

    if not isinstance(meta_model.epochs, int) or meta_model.epochs <= 0:
        raise InvalidArgumentError("epochs should be a positive integer.")

    if not isinstance(meta_model.batch_size, int) or meta_model.batch_size <= 0:
        raise InvalidArgumentError("batch_size should be a positive integer.")

    if not isinstance(meta_model.t_max, float) and not isinstance(meta_model.t_max, int):
        raise InvalidArgumentError("t_max should be float/int.")

    if meta_model.t_max <= meta_model.w_s:
        raise InvalidArgumentError(f"t_max should be greater than w_s {meta_model.w_s}.")

    if not isinstance(meta_model.delta, int) or meta_model.delta <= 0:
        raise InvalidArgumentError("delta should be a positive integer.")

    if not (0 < meta_model.s_h < 1):
        raise InvalidArgumentError("s_h should be a float between 0 and 1 exclusive.")

    if not isinstance(meta_model.epsilon, float) or meta_model.epsilon <= 0:
        raise InvalidArgumentError("epsilon should be a positive number.")

    if not isinstance(meta_model.pa, int) or meta_model.pa <= 0:
        raise InvalidArgumentError("pa should be a positive integer.")

    if not isinstance(meta_model.t_r, float) or not (0 < meta_model.t_r <= 1):
        raise InvalidArgumentError("t_r should be a positive number between 0 and 1.")

    if meta_model.time_window is not None and (
            not isinstance(meta_model.time_window, int) or meta_model.time_window <= 0):
        raise InvalidArgumentError("time_window should be a positive integer.")

    if not all(isinstance(model, (str, tuple)) and
               (isinstance(model, str) or (isinstance(model, tuple) and len(model) == 2 and isinstance(model[0],
                                                                                                       str) and
                                           isinstance(model[1], int)))
               for model in meta_model.models):
        raise InvalidArgumentError(
            "Invalid format for models. It should be a list of strings or tuples of (string, integer).")

    for model in meta_model.models:
        if isinstance(model, str):
            if model not in [LSTM, GRU, CNN, RNN_BIDIRECTIONAL, CONV_LSTM1D, RNN_ENCODER_DECODER, CNN_RNN,
                             SELF_ATTENTION, FFN]:
                raise InvalidArgumentError(
                    f"Invalid model type {model}. Supported models are {LSTM}, {GRU}, {CNN}, {RNN_BIDIRECTIONAL},"
                    f" {CONV_LSTM1D}, {RNN_ENCODER_DECODER}, {CNN_RNN}, {SELF_ATTENTION}, and {FFN}.")
        elif isinstance(model, tuple) and len(model) == 2:
            model_type, model_instances = model
            if model_type not in [LSTM, GRU, CNN, RNN_BIDIRECTIONAL, CONV_LSTM1D, RNN_ENCODER_DECODER, CNN_RNN,
                                  SELF_ATTENTION, FFN]:
                raise InvalidArgumentError(
                    f"Invalid model type {model_type}.Supported models are {LSTM}, {GRU}, {CNN}, "
                    f"{RNN_BIDIRECTIONAL},"
                    f" {CONV_LSTM1D}, {RNN_ENCODER_DECODER}, {CNN_RNN}, {SELF_ATTENTION}, and {FFN}.")
            if not isinstance(model_instances, int) or model_instances <= 0:
                raise InvalidArgumentError("Number of model instances should be a positive integer.")
        else:
            raise InvalidArgumentError(f"Invalid model specification {model}.")

    if meta_model.type_training not in [TYPE_TRAINING_AVERAGE, TYPE_TRAINING_FFN]:
        raise InvalidArgumentError(
            f"Invalid type of training technique. Supported techniques are "
            f"{TYPE_TRAINING_AVERAGE} and {TYPE_TRAINING_FFN}.")

    if meta_model.scaler not in [MIN_MAX_SCALER, STANDARD_SCALER, ROBUST_SCALER]:
        raise InvalidArgumentError(
            f"Invalid type of scaler technique. Supported techniques are {MIN_MAX_SCALER},"
            f" {STANDARD_SCALER} and {ROBUST_SCALER}.")

    if not isinstance(meta_model.use_kfold, bool):
        raise InvalidArgumentError("Invalid use_kfold parameter: must be a boolean.")

    if not 0 < meta_model.test_size < 1 or not isinstance(meta_model.test_size, float):
        raise InvalidArgumentError("Invalid test_size parameter: must be a float between 0 and 1.")

    if not 0 < meta_model.val_size < 1 or not isinstance(meta_model.val_size, float):
        raise InvalidArgumentError("Invalid val_size parameter: must be a float between 0 and 1.")

    if len(meta_model.hyperparams_ffn) != 3:
        raise ValueError("hyperparams_ffn must be a tuple of length 3")
    if len(meta_model.hyperparams_cnn) != 5:
        raise ValueError("hyperparams_cnn must be a tuple of length 5")
    if len(meta_model.hyperparams_rnn) != 3:
        raise ValueError("hyperparams_rnn must be a tuple of length 3")

    if not all(isinstance(val, int) for val in meta_model.hyperparams_ffn):
        raise ValueError("hyperparams_ffn values must be integers")
    if not all(isinstance(val, int) for val in meta_model.hyperparams_cnn):
        raise ValueError("hyperparams_cnn values must be integers")
    if not all(isinstance(val, int) for val in meta_model.hyperparams_rnn):
        raise ValueError("hyperparams_rnn values must be integers")

    if not all(val > 0 for val in meta_model.hyperparams_ffn):
        raise ValueError("hyperparams_ffn values must be greater than 0")
    if not all(val > 0 for val in meta_model.hyperparams_cnn):
        raise ValueError("hyperparams_cnn values must be greater than 0")
    if not all(val > 0 for val in meta_model.hyperparams_rnn):
        raise ValueError("hyperparams_rnn values must be greater than 0")

    if meta_model.hyperparams_ffn[1] >= meta_model.hyperparams_ffn[2]:
        raise ValueError(
            "Minimum number of neurons per layer must be less than the maximum number for hyperparams_ffn")
    if meta_model.hyperparams_cnn[0] >= meta_model.hyperparams_cnn[1]:
        raise ValueError("Minimum number of filters must be less than the maximum number for hyperparams_cnn")
    if meta_model.hyperparams_cnn[2] >= meta_model.hyperparams_cnn[3]:
        raise ValueError("Minimum kernel size must be less than the maximum kernel size for hyperparams_cnn")
    if meta_model.hyperparams_rnn[1] >= meta_model.hyperparams_rnn[2]:
        raise ValueError("Minimum number of hidden units must be less than the maximum number for hyperparams_rnn")

    if len(meta_model.hyperparams_mm_network) != 2:
        raise ValueError("hyperparams_mm_network must be a tuple of length 2")

    if not all(isinstance(val, int) for val in meta_model.hyperparams_mm_network):
        raise ValueError("hyperparams_mm_network values must be integers")

    if not isinstance(meta_model.use_multiprocessing, bool):
        raise InvalidArgumentError("Invalid use_multiprocessing parameter: must be a boolean.")

    if not isinstance(meta_model.save_models_as_dot_format, bool):
        raise InvalidArgumentError("Invalid save_models_as_dot_format parameter: must be a boolean.")



import os
from typing import Dict, Tuple

import joblib
import numpy as np
import tensorflow as tf
from numpy import ndarray
from sklearn.model_selection import KFold, train_test_split
from sklearn.preprocessing import StandardScaler

from eventdetector_ts import MODELS_DIR, META_MODEL_NETWORK, config_dict, TYPE_TRAINING_FFN, SCALERS_DIR, \
    META_MODEL_SCALER
from eventdetector_ts.metamodel.utils import DataSplitter
from eventdetector_ts.models import logger_models
from eventdetector_ts.models.helpers import CustomEarlyStopping, custom_cross_val_score
from eventdetector_ts.models.models_builder import ModelBuilder


class ModelTrainer:
    """
    A class used to train and evaluate machine learning models.

    Attributes:
        data_splitter (DataSplitter): An object of the DataSplitter class, which is used to split the data
            into train and test sets.
        epochs (int): The number of epochs to train the models.
        batch_size (int): The batch size to use during training.
        pa (int): The patience value to use for the EarlyStopping callback.
        t_r (float): The ratio value to use for the CustomEarlyStopping callback.
        use_kfold (bool): Whether to use K-Fold cross-validation or not.
        val_size (float): The size of the validation set to use during training.
        epsilon (float): A small constant used to control the size of set which contains the top models
            with the lowest MSE values.
        save_models_as_dot_format (bool): Whether to save the models as a dot format file.
                The default value is False. If set to True, then you should have graphviz software
                to be installed on your machine.
        train_losses (Dict[str, np.ndarray]): A dictionary containing the training losses for each model.
        val_losses (Dict[str, np.ndarray]): A dictionary containing the validation losses for each model.
        val_loss_meta_model (np.ndarray): val loss for the meta_model.
        train_loss_meta_model (np.ndarray): train loss for the meta_model
    """

    def __init__(self, data_splitter: DataSplitter, epochs: int,
                 batch_size: int, pa: int, t_r: float,
                 use_kfold: bool, val_size: float, epsilon: float, save_models_as_dot_format: bool) -> None:
        """
        Initialize the ModelTrainer object.

        Args:
            data_splitter (DataSplitter): An object of the DataSplitter class, which is used to split the data
                into train and test sets.
            epochs (int): The number of epochs to train the models.
            batch_size (int): The batch size to use during training.
            pa (int): The patience value to use for the EarlyStopping callback.
            t_r (float): The ratio value to use for the CustomEarlyStopping callback.
            use_kfold (bool): Whether to use K-Fold cross-validation or not.
            val_size (float): The size of the validation set to use during training.
            epsilon (float): A small constant used to control the size of set which contains the top models
                    with the lowest MSE values.
            save_models_as_dot_format (bool): Whether to save the models as a dot format file.
                The default value is False. If set to True, then you should have graphviz software
                to be installed on your machine.
        """

        self.val_loss_meta_model: list = []
        self.train_loss_meta_model: list = []
        self.save_models_as_dot_format = save_models_as_dot_format
        self.best_models: Dict[str, tf.keras.Model] = {}
        self.train_losses: Dict[str, list] = {}
        self.val_losses: Dict[str, list] = {}
        self.data_splitter: DataSplitter = data_splitter
        self.epochs: int = epochs
        self.batch_size: int = batch_size
        self.pa = pa
        self.t_r = t_r
        self.use_kfold = use_kfold
        self.val_size = val_size
        self.epsilon = epsilon

    def fitting_models(self, created_models: Dict[str, tf.keras.Model]) -> None:
        """
        Fits the created models to the training data and saves the training and validation losses.

        Args:
            created_models: A dictionary containing the created models with their names as keys
                and the models as values.

        Returns:
            None
        """
        # Define early stopping based on validation loss
        early_stopping = tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=self.pa * 2)
        # Define custom early stopping based on a ratio and patience
        custom_early_stopping = CustomEarlyStopping(ratio=self.t_r, patience=self.pa, verbose=1)
        # Loop through each model in the created models dictionary
        for model_name, model in created_models.items():
            # If using k-fold cross-validation
            if self.use_kfold:
                logger_models.info("Performing a KFold cross-validation")
                # Calculate cross validation score using custom function
                score: np.ndarray = custom_cross_val_score(model=model, x=self.data_splitter.train_x,
                                                           y=self.data_splitter.train_y,
                                                           cv=KFold(n_splits=5, shuffle=False), epochs=self.epochs,
                                                           batch_size=self.batch_size,
                                                           callbacks=[early_stopping, custom_early_stopping])
                # Print cross validation score for the current model
                logger_models.info(f"The cross validation score for {model_name} is {score}")
            # Split training data into training and validation sets
            train_x, val_x, train_y, val_y = train_test_split(self.data_splitter.train_x, self.data_splitter.train_y,
                                                              test_size=self.val_size,
                                                              shuffle=False)
            # Print message indicating fitting of a current model
            logger_models.info(f"Fitting of {model_name}...")
            # Fit the model using training data and validate using validation data
            history = model.fit(train_x, train_y, epochs=self.epochs,
                                batch_size=self.batch_size, verbose=1,
                                validation_data=(val_x, val_y),
                                callbacks=[early_stopping, custom_early_stopping])
            # Save training and validation errors for the current model
            self.train_losses[model_name] = history.history['loss']
            self.val_losses[model_name] = history.history['val_loss']

        losses_test_data: Dict[str, tf.keras.Model] = {}
        min_loss = np.Inf
        for model_name, model in created_models.items():
            logger_models.info(f"Evaluating model {model_name} on test data")
            loss = model.evaluate(self.data_splitter.test_x, self.data_splitter.test_y, batch_size=self.batch_size,
                                  use_multiprocessing=True)
            logger_models.info(f"The loss value of model {model_name} on test data is {loss:.4f}")
            losses_test_data[model_name] = loss
            if min_loss > loss:
                min_loss = loss

        logger_models.info(f"Selecting best models based on the min MSE {min_loss:.4f} and epsilon {self.epsilon}:")
        for model_name, loss_ in losses_test_data.items():
            if loss_ <= (min_loss + self.epsilon):
                self.best_models[model_name] = created_models[model_name]
        logger_models.info(f"Best models selected: {self.best_models.keys()}")

        config_dict["models"] = list(self.best_models.keys())

    def save_best_models(self, output_dir: str) -> None:
        """
        Save the best models to the specified output directory.

        Args:
             output_dir (str): The directory to save the best models.

        Returns:
            None
        """

        for model_name, model in self.best_models.items():
            # Print the name of the current model being saved
            logger_models.info(f"Current model to be saved on the disk is {model_name}")
            path = os.path.join(output_dir, MODELS_DIR)
            # Save the model to the specified directory
            model_path = os.path.join(path, model_name)
            model.save(model_path)
        logger_models.info("Models saved successfully.")

    def train_meta_model(self, type_training: str, hyperparams_mm_network: Tuple[int, int], output_dir: str) \
            -> tuple[ndarray, float, ndarray]:
        """
        Trains the metamodel using the best models predictions as features.

         Args:
            type_training: The type of training to use, either "ffn" or "mean".
            hyperparams_mm_network: A tuple containing the hyperparameters the MetaModel network.
            output_dir: The directory to save the trained models to.

        Returns:
            A tuple containing the final prediction and the loss.
        """
        predictions = []
        for model_name, model in self.best_models.items():
            # Make predictions for the test set using each model
            predicted_y: np.ndarray = model.predict(self.data_splitter.test_x, batch_size=self.batch_size,
                                                    use_multiprocessing=True)
            predicted_y = predicted_y.flatten()
            predictions.append(predicted_y)

        # Convert a list of 1D NumPy arrays to 2D NumPy array
        x = np.stack(predictions, axis=1)

        if type_training == TYPE_TRAINING_FFN:
            logger_models.info("Train the MetaModel using a FFN to produce a final prediction")
            # Split the data into training and test sets
            train_x, test_x, train_y, test_y = train_test_split(x, self.data_splitter.test_y,
                                                                test_size=self.data_splitter.test_size,
                                                                shuffle=False)
            scaler = StandardScaler()
            train_x = scaler.fit_transform(train_x)
            test_x = scaler.transform(test_x)
            scalers_dir = os.path.join(output_dir, SCALERS_DIR)
            scaler_path = os.path.join(scalers_dir, f"{META_MODEL_SCALER}.joblib")
            joblib.dump(scaler, scaler_path)
            # Build the FFN model
            inputs = tf.keras.Input(shape=(train_x.shape[1],), name="input")
            layers, units = hyperparams_mm_network
            model_builder: ModelBuilder = ModelBuilder(inputs=inputs)

            for _ in range(layers):
                units_j = units
                model_builder.add_dense_layer(units=units_j)
            model_builder.add_dense_layer(units=1, dropout=None)
            keras_model = model_builder.build(name=META_MODEL_NETWORK, root_dir=output_dir,
                                              save_models_as_dot_format=self.save_models_as_dot_format)
            # Train the model
            logger_models.info("Fitting the MetaModel network...")
            history = keras_model.fit(train_x, train_y, epochs=self.epochs, batch_size=self.batch_size, verbose=1,
                                      validation_data=(test_x, test_y))

            path = os.path.join(output_dir, MODELS_DIR)
            model_path = os.path.join(path, META_MODEL_NETWORK)
            keras_model.save(model_path)
            logger_models.info("MetaModel network saved successfully.")
            self.train_loss_meta_model = history.history['loss']
            self.val_loss_meta_model = history.history['val_loss']

            # final_prediction: np.ndarray = keras_model.predict(self.data_splitter.test_x, batch_size=self.batch_size,
            #                                                   use_multiprocessing=True)
            final_prediction: np.ndarray = keras_model.predict(test_x, batch_size=self.batch_size,
                                                               use_multiprocessing=False)
            final_prediction = final_prediction.flatten()
            return final_prediction, tf.keras.losses.mse(final_prediction, test_y), test_y
        else:
            # Compute the average prediction
            logger_models.info("Compute the average of predictions to produce a final prediction")
            final_prediction = np.mean(x, axis=1)
            return final_prediction, tf.keras.losses.mse(final_prediction,
                                                         self.data_splitter.test_y), self.data_splitter.test_y



import json
import os
from typing import Dict, List, Tuple, Any

import joblib
import numpy as np
import pandas as pd
import tensorflow as tf

from eventdetector_ts import CONFIG_FILE, SCALERS_DIR, TYPE_TRAINING_FFN, TimeUnit, MODELS_DIR, META_MODEL_NETWORK, \
    META_MODEL_SCALER
from eventdetector_ts.data.helpers import convert_dataframe_to_overlapping_partitions, get_timedelta
from eventdetector_ts.optimization.algorithms import convolve_with_gaussian_kernel
from eventdetector_ts.optimization.event_extraction_pipeline import get_peaks, compute_op_as_mid_times
from eventdetector_ts.prediction import logger


def load_config_file(path: str) -> Dict:
    """
     Load config file of the meta-model.
     
    Args:
        path (str): Where the config file is stored

    Returns:
        Data as a Dict which contains all configuration information
    """
    config_file_path = os.path.join(path, CONFIG_FILE)
    if not os.path.exists(config_file_path):
        msg: str = f"The config file {CONFIG_FILE} does not exist in this path: {config_file_path}"
        logger.critical(msg)
        raise ValueError(msg)

    with open(config_file_path, 'r') as f:
        config_: Dict = json.load(f)
        return config_


def load_models(model_keys: List[str], output_dir: str) -> List[tf.keras.Model]:
    """
    Loads the trained models.
    Args:
        model_keys (List[str]): List of model's name
        output_dir (str): The parent directory where the trained models are stored

    Returns:
        List of keras models
    """
    models: List[tf.keras.Model] = []
    for key in model_keys:
        path = os.path.join(output_dir, MODELS_DIR)
        path = os.path.join(path, key)
        models.append(tf.keras.models.load_model(path))
    return models


def apply_scaling(x: np.ndarray, config_data: Dict) -> np.ndarray:
    """
    Scaling input data according to the stored scalers.
    Args:
        x (np.ndarray): Input data to be scaled 
        config_data (Dict): Configuration Data 

    Returns:
        Scaled data.
    """
    n_time_steps = x.shape[1]
    output_dir: str = config_data.get("output_dir")
    scalers_dir = os.path.join(output_dir, SCALERS_DIR)
    try:
        for i in range(n_time_steps):
            scaler_i_path = os.path.join(scalers_dir, f'scaler_{i}.joblib')
            # Print progress
            print("\rLoading and applying scalers...{}/{}".format(i + 1, n_time_steps), end="")
            # Load the scaler from disk
            scaler = joblib.load(scaler_i_path)
            x[:, i, :] = scaler.transform(x[:, i, :])
    except ValueError as e:
        logger.critical(e)
        raise e

    logger.info("Convert data to float32 for consistency...")
    x = np.asarray(x).astype('float32')
    return x


def load_meta_model(output_dir: str) -> Tuple[tf.keras.Model, Any]:
    """
    Load the metamodel network and the scaler.
    Args:
        output_dir (str): The parent directory where the trained models are stored

    Returns:
        tf.keras.Model, StanderScaler
    """
    path = os.path.join(output_dir, MODELS_DIR)
    path = os.path.join(path, META_MODEL_NETWORK)
    model = tf.keras.models.load_model(path)
    scalers_dir = os.path.join(output_dir, SCALERS_DIR)
    scaler_path = os.path.join(scalers_dir, f'{META_MODEL_SCALER}.joblib')
    scaler = joblib.load(scaler_path)

    return model, scaler


def predict(dataset: pd.DataFrame, path: str) -> Tuple[List, np.ndarray, np.ndarray]:
    """
    Generates output predictions for the input dataset
    Args:
        dataset (pd.DataFrame): The input dataset.
        path (str): The path to the created folder by the MetaModel. 

    Returns:
        Tuple[List, np.ndarray, np.ndarray]: Predicted events, predicted Op and filtered predicted Op
    """

    if path is None or not isinstance(path, str) or len(path) == 0:
        msg: str = f"The provided path {path} is not valid."
        logger.critical(msg)
        raise ValueError(msg)

    config_data: Dict = load_config_file(path=path)
    config_data['output_dir'] = path
    logger.info(f"Config dict: {config_data}")
    logger.info("Converting the dataset to overalapping partitions.")
    dataset_as_overlapping_partitions: np.ndarray = convert_dataframe_to_overlapping_partitions(dataset,
                                                                                                width=config_data.get(
                                                                                                    "width"),
                                                                                                step=config_data.get(
                                                                                                    "step"),
                                                                                                fill_method=config_data.get(
                                                                                                    'fill_nan'))
    # Remove the column containing the timestamps from the overalapping partitions
    x: np.ndarray = np.delete(dataset_as_overlapping_partitions, -1, axis=2)
    logger.info(f"The shape of the input data: {x.shape}")
    x = apply_scaling(x=x, config_data=config_data)
    model_keys: List[str] = config_data.get('models')
    logger.info(f"Loading models: {model_keys}")
    models: List[tf.keras.Model] = load_models(model_keys=model_keys, output_dir=config_data.get('output_dir'))
    batch_size: int = config_data.get("batch_size")
    predictions = []
    logger.info("Making prediction from the trained models")
    for model in models:
        # Make predictions using each model
        predicted_y: np.ndarray = model.predict(x, batch_size=batch_size,
                                                use_multiprocessing=False)
        predicted_y = predicted_y.flatten()
        predictions.append(predicted_y)

    type_training: str = config_data.get('type_training')
    # Convert a list of 1D NumPy arrays to 2D NumPy array
    predictions = np.stack(predictions, axis=1)
    if type_training == TYPE_TRAINING_FFN:
        logger.info("Loading the MetaModel and its Scaler")
        model, scaler = load_meta_model(output_dir=config_data.get('output_dir'))
        predictions = scaler.transform(predictions)
        logger.info("Make a final prediction using the network of the MetaModel")
        predicted_op = model.predict(predictions, batch_size=batch_size)
        predicted_op = predicted_op.flatten()
    else:
        logger.info("Make a final prediction by averaging")
        predicted_op = np.mean(predictions, axis=1)

    sigma, m, h = config_data.get('best_combination')
    logger.info(f"Applying Gaussian Filter with sigma = {sigma} and m = {m}")
    filtered_predicted_op = convolve_with_gaussian_kernel(predicted_op, sigma=sigma, m=m)
    logger.info("Computing filtered predictions as a function of the mid-times of the overalapping partitions")
    t, filtered_predicted_op = compute_op_as_mid_times(overlapping_partitions=dataset_as_overlapping_partitions,
                                                       op_g=filtered_predicted_op)
    logger.info(f"Computing peaks with h = {h:.2f}")
    s_peaks = get_peaks(h=h, t=t, op_g=filtered_predicted_op)
    predicted_events = []
    time_unit: TimeUnit = TimeUnit.__call__(config_data.get('time_unit'))
    radius = get_timedelta(config_data.get("w_s") // 2, time_unit)
    logger.info(f"Generating a predicted events with radius = {radius}, predicted op and a filtered predicted op")
    for i in range(len(s_peaks)):
        predicted_event = s_peaks[i]
        start_time = predicted_event - radius
        end_time = predicted_event + radius
        predicted_events.append((start_time.isoformat(), end_time.isoformat()))
    return predicted_events, predicted_op, filtered_predicted_op



import os
import random
from typing import Optional, Tuple, Dict, List

import numpy as np
import tensorflow as tf

from eventdetector_ts import LSTM, GRU, CNN, RNN_BIDIRECTIONAL, RNN_ENCODER_DECODER, CNN_RNN, FFN, CONV_LSTM1D, \
    SELF_ATTENTION, MODELS_DIR
from eventdetector_ts.models import logger_models
from eventdetector_ts.models.helpers import SelfAttention


class DtwLoss(tf.keras.losses.Loss):
    def __init__(self, batch_size: int = 32):
        super(DtwLoss, self).__init__()
        self.batch_size = batch_size

    def call(self, y_true, y_pred):
        tmp = []
        for item in range(self.batch_size):
            s = y_true[item, :]
            t = y_pred[item, :]
            n, m = len(s), len(t)
            dtw_matrix = []
            for i in range(n + 1):
                line = []
                for j in range(m + 1):
                    if i == 0 and j == 0:
                        line.append(0)
                    else:
                        line.append(np.inf)
                dtw_matrix.append(line)

            for i in range(1, n + 1):
                for j in range(1, m + 1):
                    cost = tf.abs(s[i - 1] - t[j - 1])
                    last_min = tf.reduce_min([dtw_matrix[i - 1][j], dtw_matrix[i][j - 1], dtw_matrix[i - 1][j - 1]])
                    dtw_matrix[i][j] = tf.cast(cost, dtype=tf.float32) + tf.cast(last_min, dtype=tf.float32)

            temp = []
            for i in range(len(dtw_matrix)):
                temp.append(tf.stack(dtw_matrix[i]))

            tmp.append(tf.stack(temp)[n, m])
        return tf.reduce_mean(tmp)


class ModelBuilder:
    """
    Helper class for building TensorFlow Keras models.
    """

    def __init__(self, inputs):
        """
        Initialize the ModelBuilder object.

        Args:
            inputs (tf.keras.layers.Layer): The input layer of the model.
        """
        self.outputs = None
        self.inputs = inputs

    def __return_input(self):
        """
        Return the input layer of the model if output is None, otherwise return the output layer.

        Returns:
            tf.keras.layers.Layer: The input or output layer of the model.
        """
        if self.outputs is None:
            return self.inputs
        return self.outputs

    def __check_input_shape(self, layer: tf.keras.layers.Layer) -> tf.Tensor:
        """
        Check if the input shape of a layer is compatible with the output shape of the previous layer.

        Example:
            input_1 = tf.keras.layers.Input(shape=(10,))
            input_2 = tf.keras.layers.Input(shape=(20,))
            concat_layer = tf.keras.layers.Concatenate()([input_1, input_2])
            print(concat_layer.input_spec) -> [InputSpec(shape=(None, 10), dtype=tf.float32),
                                                InputSpec(shape=(None, 20), dtype=tf.float32)]

        Args:
            layer: A `tf.keras.layers.Layer` object for which to check the input shape.

        Returns:
            A `tf.Tensor` object representing the output of the previous layer (i.e., the input to the current layer).

        Raises:
            ValueError: If the input shape is not compatible with the output shape of the previous layer.
        """
        input_ = self.__return_input()

        if not hasattr(layer, 'input_spec') or layer.input_spec is None:
            return input_

        input_spec = layer.input_spec
        if not isinstance(input_spec, list):
            input_spec = [input_spec]

        for spec in input_spec:
            # Retrieve the expected shape and dtype from the input spec
            expected_shape = spec.shape
            expected_dtype = spec.dtype

            # Check if the expected shape is defined
            if expected_shape is not None:
                expected_rank = len(expected_shape)
            else:
                # If the expected shape is not defined, set the expected rank to None
                expected_rank = None

            # Get the actual input shape, dtype, and rank
            actual_shape = input_.shape[1:]
            actual_dtype = input_.dtype
            actual_rank = len(actual_shape)

            # Check if the shape, dtype, and rank are compatible
            if expected_shape is not None and expected_shape != actual_shape:
                msg = f"The expected input shape for layer '{layer.name}' is {expected_shape}, " \
                      f"but received {actual_shape} instead."
                logger_models.critical(msg)
                raise ValueError(msg)
            if expected_dtype is not None and expected_dtype != actual_dtype:
                msg = f"The expected input dtype for layer '{layer.name}' is {expected_dtype}, " \
                      f"but received {actual_dtype} instead."
                logger_models.critical(msg)
                raise ValueError(msg)
            if expected_rank is not None and expected_rank != actual_rank:
                msg = f"The expected input rank for layer '{layer.name}' is {expected_rank}, " \
                      f"but received {actual_rank} instead."
                logger_models.critical(msg)
                raise ValueError(msg)

        return input_

    def __add_layer(self, layer: tf.keras.layers.Layer, check_shape: bool = True) -> None:
        """
        Adds a layer to the model and checks if the layer's input shape is compatible with the previous
        layer's output shape.

        Args:
            layer: The layer to be added.
            check_shape: A boolean indicating whether to check the input shape compatibility or not.

        Returns:
            None.

        Raises:
            ValueError: If the input shape is not compatible with the output shape of the previous layer.
        """

        if check_shape:
            input_ = self.__check_input_shape(layer)
        else:
            input_ = self.__return_input()

        # Add the layer to the model by passing its input as input_ and saving the output to self.outputs
        self.outputs = layer(input_)

    def add_lstm_layer(self, hidden_dim: int, activation: str = "tanh", return_sequences: bool = False,
                       dropout: float = 0.3) -> None:
        """
        Adds a Long-Short Term Memory (LSTM) layer to the model.

        Args:
            hidden_dim (int): The number of output units in the LSTM layer.
            activation (str, optional): The activation function to use. Defaults to 'tanh'.
            return_sequences (bool, optional): Whether to return the last output in the output sequence
                or the full sequence. Default to False.
            dropout (float, optional): The dropout rate to use. Default to 0.3.

        Returns:
            None
        """

        lstm = tf.keras.layers.LSTM(units=hidden_dim, activation=activation, return_sequences=return_sequences,
                                    dropout=dropout)
        self.__add_layer(lstm)

    def add_conv1d_layer(self, filters: int, kernel_size: int, strides: int = 1, padding: str = 'same',
                         activation: str = "relu") -> None:
        """
        Adds a 1D convolutional layer to the model.

        Args:
            filters (int): The number of output filters in the convolution.
            kernel_size (int): The size of the convolutional kernel.
            strides (int): The stride of the convolution. Default is 1.
            padding (str): One of 'valid' or 'same'. Default is 'same'.
            activation (str): The activation function to use. Default is 'relu'.

        Returns:
            None
        """

        conv1d_layer = tf.keras.layers.Conv1D(filters=filters, kernel_size=kernel_size, strides=strides,
                                              padding=padding, activation=activation)
        self.__add_layer(conv1d_layer)

    def add_dense_layer(self, units: int, activation: str = 'sigmoid', dropout: Optional[float] = 0.3) -> None:
        """
        Add a dense layer to the neural network.

        Args:
            units (int): The number of units in the dense layer.
            activation (str): The activation function to use in the dense layer. Default is 'sigmoid'.
            dropout (Optional[float]): The dropout rate to use in the dense layer. If set to None, no dropout
                layer will be added. Default is 0.3.

        Returns:
            None
        """

        dense_layer = tf.keras.layers.Dense(units=units, activation=activation)
        self.__add_layer(dense_layer)
        if dropout is not None:
            dropout_layer = tf.keras.layers.Dropout(rate=dropout)
            self.__add_layer(dropout_layer)

    def add_gru_layer(self, hidden_dim: int, activation: str = "tanh", return_sequences: bool = False,
                      dropout: float = 0.3) -> None:
        """
        Adds a Gated Recurrent Unit (GRU) layer to the model.

        Args:
            hidden_dim (int): The number of output units in the GRU layer.
            activation (str, optional): The activation function to use. Defaults to 'tanh'.
            return_sequences (bool, optional): Whether to return the last output in the output sequence
                or the full sequence. Defaults to False.
            dropout (float, optional): The dropout rate to use. Default to 0.3.

        Returns:
            None
        """

        gru_layer = tf.keras.layers.GRU(units=hidden_dim, activation=activation, return_sequences=return_sequences,
                                        dropout=dropout)
        self.__add_layer(gru_layer)

    def add_bidirectional(self, hidden_dim: int, activation: str = "tanh",
                          return_sequences: bool = False, dropout: float = 0.3) -> None:
        """
        Adds a Bidirectional LSTM layer to the model.

        Args:
            hidden_dim (int): Number of LSTM units in the layer.
            activation (str, optional): Activation function to use. Defaults to 'tanh'.
            return_sequences (bool, optional): Whether to return the last output in the output sequence,
                                                or the full sequence. Defaults to False.
            dropout (float, optional): The dropout rate to apply. Default to 0.3.

        Returns:
            None
        """

        lstm_fw = tf.keras.layers.LSTM(units=hidden_dim, activation=activation,
                                       return_sequences=return_sequences, dropout=dropout)
        lstm_bw = tf.keras.layers.LSTM(units=hidden_dim, activation=activation,
                                       return_sequences=return_sequences, dropout=dropout, go_backwards=True)
        bidirectional = tf.keras.layers.Bidirectional(lstm_fw, backward_layer=lstm_bw, merge_mode="mul")
        self.__add_layer(bidirectional)

    def add_conv_lstm1d(self, filters: int, kernel_size: int, activation: str = 'tanh',
                        strides: int = 1, padding: str = 'same', dropout: float = 0.3,
                        return_sequences: bool = False) -> None:
        """
        Adds a ConvLSTM1D layer to the neural network.

        Args:
            filters (int): Number of output filters in the convolution.
            kernel_size (int): Size of the 1D convolutional partition.
            activation (str): Activation function to use. Defaults to 'tanh'.
            strides (int): Stride length of the convolution. Default to 1.
            padding (str): Type of padding to use. Defaults to 'same'.
            dropout (float): Fraction of the input units to drop. Default to 0.3.
            return_sequences (bool, optional): Whether to return the last output in the output sequence,
                                                or the full sequence. Defaults to False.

        Returns:
            None
        """

        conv_lstm1d_layer = tf.keras.layers.ConvLSTM1D(filters=filters, kernel_size=kernel_size, activation=activation,
                                                       strides=strides, padding=padding, dropout=dropout,
                                                       return_sequences=return_sequences)
        self.__add_layer(conv_lstm1d_layer)

    def add_flatten_layer(self) -> None:
        """
        Add flatten layer

        Returns:
            None
        """

        flatten = tf.keras.layers.Flatten()
        self.__add_layer(flatten)

    def add_max_pooling1d(self, pool_size: int, strides: Optional[int] = None) -> None:
        """
        Adds a 1D max pooling layer to the model.

        Args:
            pool_size (int): Size of the max pooling partitions.
            strides (int, optional): Factor by which to downscale. Defaults to None.

        Returns:
            None
        """

        # Create a 1D max pooling layer with the given pool_size and strides
        pooling = tf.keras.layers.MaxPooling1D(pool_size=pool_size, strides=strides)
        # Add the max pooling layer to the model
        self.__add_layer(pooling)

    def add_reshape(self, shape: tuple) -> None:
        """
        Adds a reshape layer to the model.

        Args:
            shape (tuple): A tuple representing the new shape for the input tensor.

        Returns:
            None
        """

        # Create a reshaped layer with the given shape
        reshape = tf.keras.layers.Reshape(target_shape=shape)
        # Add the reshaped layer to the model
        self.__add_layer(reshape)

    def add_self_attention(self, units: int) -> None:
        """
        Adds a self-attention layer to the model.

        Args:
            units (int): Dimensionality of the output space.

        Returns:
            None
        """

        # Create a self-attention layer with the given units
        self_attention_layer = SelfAttention(units)
        # Add the self-attention layer to the model
        self.__add_layer(self_attention_layer)

    def add_global_max_pooling(self):
        """
        Adds a global max pooling layer to the model.

        Returns:
            None
        """
        # Create a global max pooling layer
        pooling = tf.keras.layers.GlobalMaxPooling1D()
        # Add the global max pooling layer to the model
        self.__add_layer(pooling)

    def build(self, name: str, save_models_as_dot_format: bool, root_dir: Optional[str] = None,
              ) -> tf.keras.Model:
        """
        Builds the model and compiles it.

        Args:
            name: The name of the model.
            save_models_as_dot_format: Whether to save the model as a dot format file.
            root_dir (Optional[str]): The root directory in which to save the model as a dot format file.

        Returns:
            The compiled model.
        """
        model_ = tf.keras.Model(inputs=self.inputs, outputs=self.outputs, name=name)

        if save_models_as_dot_format and root_dir is not None:
            png_path = os.path.join(root_dir, f"{name}.png")
            tf.keras.utils.plot_model(model_, png_path, show_shapes=True, show_dtype=True,
                                      show_layer_names=True, expand_nested=True)

        model_.compile(
            optimizer=tf.keras.optimizers.Adam(1e-3),
            loss=tf.keras.losses.MeanSquaredError()
        )

        return model_


class ModelCreator:
    """
       A class for creating different neural network models.

       Attributes:
           models (List[Tuple[str, int]]): A list of tuples representing the model types and number of instances.
               Supported model types include LSTM, GRU, CNN, RNN_BIDIRECTIONAL, CONV_LSTM1D, RNN_ENCODER_DECODER,
               CNN_RNN, SELF_ATTENTION, and FFN.
           hyperparams_ffn (Tuple[int, int, int]): Specify the hyper-parameters for the FFN.
           hyperparams_cnn (Tuple[int, int, int, int, int]): Specify the hyper-parameters for the CNN.
           hyperparams_rnn (Tuple[int, int, int]): Specify the hyper-parameters for the RNN.
           inputs (tf.keras.Input): The input layer for the neural network.
           save_models_as_dot_format (bool): Whether to save the models as a dot format file.
           root_dir (str): The root directory where the created models will be saved.

       """

    def __init__(self, models: List[Tuple[str, int]], hyperparams_ffn: Tuple[int, int, int],
                 hyperparams_cnn: Tuple[int, int, int, int, int], hyperparams_rnn: Tuple[int, int, int],
                 save_models_as_dot_format: bool,
                 root_dir: Optional[str] = None):
        """
        Initialize the ModelCreator class with the given arguments.

        Args:
            models (List[Tuple[str, int]]): A list of tuples representing the model types and
                number of instances. Supported model types include LSTM, GRU, CNN, RNN_BIDIRECTIONAL, CONV_LSTM1D,
                RNN_ENCODER_DECODER, CNN_RNN, SELF_ATTENTION, and FFN.
            hyperparams_ffn (Tuple[int, int, int]): Specify the hyper-parameters for the FFN.
            hyperparams_cnn (Tuple[int, int, int, int, int]): Specify the hyper-parameters for the CNN.
            hyperparams_rnn (Tuple[int, int, int]): Specify the hyperparameters for the RNN.
            save_models_as_dot_format (bool): Whether to save the models as a dot format file.
                The default value is False. If set to True, then you should have graphviz software
                to be installed on your machine.
            root_dir (str): The root directory where the created models will be saved.
        """
        self.save_models_as_dot_format = save_models_as_dot_format
        self.train_losses: Dict[str, np.ndarray] = {}
        self.val_losses: Dict[str, np.ndarray] = {}
        self.root_dir: str = root_dir
        self.inputs = None
        self.hyperparams_rnn = hyperparams_rnn
        self.hyperparams_cnn = hyperparams_cnn
        self.hyperparams_ffn = hyperparams_ffn
        self.models = models
        self.created_models: Dict[str, tf.keras.Model] = {}
        self.__create_models_dir()

    def __create_models_dir(self):
        if self.root_dir is not None:
            # Create the directory if it doesn't exist
            models_dir = os.path.join(self.root_dir, MODELS_DIR)
            if not os.path.exists(models_dir):
                os.makedirs(models_dir)
            self.root_dir = models_dir

    def __create_lstm_networks(self) -> None:
        """
        Creates LSTM networks with random number of layers and units within the specified range.

        Returns:
            None
        """
        # Get the maximum number of LSTM units and layers from hyperparameters
        max_layers, min_units, max_units = self.hyperparams_rnn
        # Get the number of LSTM instances from the model list
        num_instances = self.__get_instances(LSTM)
        # If there are LSTM instances, create them
        if num_instances > 0:
            for i in range(num_instances):
                # Create a new instance of the ModelBuilder class
                model: ModelBuilder = ModelBuilder(inputs=self.inputs)
                # Set a base name for the model
                name = f"{LSTM}_{i}"
                # Choose a random number of layers between 1 and the maximum number of layers
                num_layers = random.randint(1, max_layers + 1)
                units = [random.randint(min_units, max_units + 1) for _ in range(num_layers)]
                units = sorted(units, reverse=True)
                for j in range(num_layers):
                    units_j = units[j]
                    if j == num_layers - 1:
                        model.add_lstm_layer(hidden_dim=units_j)
                    else:
                        model.add_lstm_layer(hidden_dim=units_j, return_sequences=True)
                # adds a dense layer with a single unit to a neural network model,
                # for regression where the output is a continuous numerical value.
                model.add_dense_layer(units=1, dropout=None)
                # Build the model with the chosen name and save it to the created_models dictionary
                keras_model = model.build(name=name, save_models_as_dot_format=self.save_models_as_dot_format,
                                          root_dir=self.root_dir)
                self.created_models[name] = keras_model

    def __create_gru_networks(self) -> None:
        """
        Creates GRU networks with random number of layers and units within the specified range.

        Returns:
            None
        """
        max_layers, min_units, max_units = self.hyperparams_rnn
        num_instances = self.__get_instances(GRU)

        if num_instances > 0:
            for i in range(num_instances):
                model: ModelBuilder = ModelBuilder(inputs=self.inputs)
                name = f"{GRU}_{i}"
                num_layers = random.randint(1, max_layers + 1)
                units = [random.randint(min_units, max_units + 1) for _ in range(num_layers)]
                units = sorted(units, reverse=True)
                for j in range(num_layers):
                    units_j = units[j]
                    if j == num_layers - 1:
                        model.add_gru_layer(hidden_dim=units_j)
                    else:
                        model.add_gru_layer(hidden_dim=units_j, return_sequences=True)
                # Add last layer for regression
                model.add_dense_layer(units=1, dropout=None)
                keras_model = model.build(name=name, save_models_as_dot_format=self.save_models_as_dot_format,
                                          root_dir=self.root_dir)
                self.created_models[name] = keras_model

    def __create_cnn_networks(self):
        """
        Create convolutional neural network models with a random number of convolutional layers and filters.

        Returns:
            None
        """
        filters_min, filters_max, kernel_size_min, kernel_size_max, max_layers = self.hyperparams_cnn
        num_instances = self.__get_instances(CNN)

        if num_instances > 0:
            for i in range(num_instances):
                model: ModelBuilder = ModelBuilder(inputs=self.inputs)
                name = f"{CNN}_{i}"
                num_layers = random.randint(1, max_layers + 1)
                filters_per_layer = random.randint(filters_min, filters_max + 1)
                kernel_size_per_layer = random.randint(kernel_size_min, kernel_size_max + 1)
                for j in range(num_layers):
                    filters = max(1, filters_per_layer // (2 ** (num_layers - j - 1)))
                    kernel_size = kernel_size_per_layer
                    model.add_conv1d_layer(filters=filters, kernel_size=kernel_size)
                    if j % 2 == 1:
                        model.add_max_pooling1d(pool_size=2, strides=2)

                # Add flatten layer before dense layer
                model.add_flatten_layer()
                # Add dense output layer for regression
                model.add_dense_layer(units=1, dropout=None)
                keras_model = model.build(name=name, save_models_as_dot_format=self.save_models_as_dot_format,
                                          root_dir=self.root_dir)
                self.created_models[name] = keras_model

    def __create_bi_lstm(self):
        max_layers, min_units, max_units = self.hyperparams_rnn
        num_instances = self.__get_instances(RNN_BIDIRECTIONAL)
        if num_instances > 0:
            for i in range(num_instances):
                model: ModelBuilder = ModelBuilder(inputs=self.inputs)
                name = f"{RNN_BIDIRECTIONAL}_{i}"
                num_layers = random.randint(1, max_layers + 1)
                units = [random.randint(min_units, max_units + 1) for _ in range(num_layers)]
                units = sorted(units, reverse=True)
                for j in range(num_layers):
                    units_j = units[j]
                    if j == num_layers - 1:
                        model.add_bidirectional(hidden_dim=units_j)
                    else:
                        model.add_bidirectional(hidden_dim=units_j, return_sequences=True)
                # Add last layer for regression
                model.add_dense_layer(units=1, dropout=None)
                keras_model = model.build(name=name, save_models_as_dot_format=self.save_models_as_dot_format,
                                          root_dir=self.root_dir)
                self.created_models[name] = keras_model

    def __create_ffn(self):
        max_layers, min_units, max_units = self.hyperparams_ffn
        num_instances = self.__get_instances(FFN)
        if num_instances > 0:
            for i in range(num_instances):
                model: ModelBuilder = ModelBuilder(inputs=self.inputs)
                model.add_flatten_layer()
                name = f"{FFN}_{i}"
                num_layers = random.randint(1, max_layers + 1)
                units = [random.randint(min_units, max_units + 1) for _ in range(num_layers)]
                units = sorted(units, reverse=True)
                for j in range(num_layers):
                    units_j = units[j]
                    model.add_dense_layer(units=units_j)
                # Add last layer for regression
                model.add_dense_layer(units=1, dropout=None)
                keras_model = model.build(name=name, save_models_as_dot_format=self.save_models_as_dot_format,
                                          root_dir=self.root_dir)
                self.created_models[name] = keras_model

    def __create_rnn_encoder_decoder(self):
        max_layers, min_units, max_units = self.hyperparams_rnn
        num_instances = self.__get_instances(RNN_ENCODER_DECODER)
        if num_instances > 0:
            for i in range(num_instances):
                model: ModelBuilder = ModelBuilder(inputs=self.inputs)
                name = f"{RNN_ENCODER_DECODER}_{i}"
                num_layers = random.randint(1, max_layers + 1)
                units = [random.randint(min_units, max_units + 1) for _ in range(num_layers)]
                units = sorted(units, reverse=True)
                for j in range(num_layers):
                    units_j = units[j]
                    if j == num_layers - 1:
                        model.add_lstm_layer(hidden_dim=units_j, return_sequences=True)
                        model.add_gru_layer(hidden_dim=units_j)
                    else:
                        model.add_lstm_layer(hidden_dim=units_j, return_sequences=True)
                        model.add_gru_layer(hidden_dim=units_j, return_sequences=True)

                # Add last layer for regression
                model.add_dense_layer(units=1, dropout=None)
                keras_model = model.build(name=name, save_models_as_dot_format=self.save_models_as_dot_format,
                                          root_dir=self.root_dir)
                self.created_models[name] = keras_model

    def __create_cnn_rnn(self):
        max_layers, min_units, max_units = self.hyperparams_rnn
        filters_max, filters_min, kernel_size_min, kernel_size_max, max_layers_ = self.hyperparams_cnn
        max_layers = max(max_layers_, max_layers)
        num_instances = self.__get_instances(CNN_RNN)
        if num_instances > 0:
            for i in range(num_instances):
                model: ModelBuilder = ModelBuilder(inputs=self.inputs)
                name = f"{CNN_RNN}_{i}"
                num_layers = random.randint(1, max_layers + 1)
                units = [random.randint(min_units, max_units + 1) for _ in range(num_layers)]
                units = sorted(units, reverse=True)
                filters_per_layer = random.randint(filters_min, filters_max + 1)
                kernel_size_per_layer = random.randint(kernel_size_min, kernel_size_max + 1)
                for j in range(num_layers):
                    units_j = units[j]
                    filters = max(1, filters_per_layer // (2 ** (num_layers - j - 1)))
                    kernel_size = kernel_size_per_layer
                    model.add_conv1d_layer(filters=filters, kernel_size=kernel_size)
                    model.add_lstm_layer(hidden_dim=units_j, return_sequences=True)
                    if j % 2 == 1:
                        model.add_max_pooling1d(pool_size=2, strides=2)

                # Add flatten layer before dense layer
                model.add_flatten_layer()
                # Add last layer for regression
                model.add_dense_layer(units=1, dropout=None)
                keras_model = model.build(name=name, save_models_as_dot_format=self.save_models_as_dot_format,
                                          root_dir=self.root_dir)
                self.created_models[name] = keras_model

    def __create_conv_lstm1d(self):
        filters_min, filters_max, kernel_size_min, kernel_size_max, max_layers = self.hyperparams_cnn
        num_instances = self.__get_instances(CONV_LSTM1D)
        if num_instances > 0:
            for i in range(num_instances):
                model: ModelBuilder = ModelBuilder(inputs=self.inputs)
                shape_inputs = tf.keras.backend.int_shape(model.inputs)
                new_shape = (shape_inputs[1], 1, shape_inputs[2])
                model.add_reshape(shape=new_shape)
                name = f"{CONV_LSTM1D}_{i}"
                num_layers = random.randint(1, max_layers + 1)
                filters_per_layer = random.randint(filters_min, filters_max + 1)
                kernel_size_per_layer = random.randint(kernel_size_min, kernel_size_max + 1)
                for j in range(num_layers):
                    filters = max(1, filters_per_layer // (2 ** (num_layers - j - 1)))
                    kernel_size = kernel_size_per_layer
                    model.add_conv_lstm1d(filters=filters, kernel_size=kernel_size, return_sequences=True)

                # Add flatten layer before dense layer
                model.add_flatten_layer()
                # Add last layer for regression
                model.add_dense_layer(units=1, dropout=None)
                keras_model = model.build(name=name, save_models_as_dot_format=self.save_models_as_dot_format,
                                          root_dir=self.root_dir)
                self.created_models[name] = keras_model

    def __create_encoder_decoder_self_attention(self):
        max_layers, min_units, max_units = self.hyperparams_rnn
        num_instances = self.__get_instances(SELF_ATTENTION)

        if num_instances > 0:
            for i in range(num_instances):
                model: ModelBuilder = ModelBuilder(inputs=self.inputs)
                name = f"{SELF_ATTENTION}_{i}"
                num_layers = random.randint(1, max_layers + 1)
                units = [random.randint(min_units, max_units + 1) for _ in range(num_layers)]
                units = sorted(units, reverse=True)
                for j in range(num_layers):
                    units_j = units[j]
                    model.add_bidirectional(hidden_dim=units_j, return_sequences=True)
                    model.add_gru_layer(hidden_dim=units_j, return_sequences=True)
                    model.add_self_attention(units=units_j)
                    if j == num_layers - 1:
                        # Two cases
                        case1 = "USE_LSTM"
                        case2 = "USE_MAX_GLOBAl_POOLING"
                        chosen_case = random.choice([case1, case2])
                        if chosen_case == case1:
                            model.add_lstm_layer(hidden_dim=units_j, return_sequences=False)
                        else:
                            model.add_global_max_pooling()
                # Add last layer for regression
                model.add_dense_layer(units=1, dropout=None)
                keras_model = model.build(name=name, save_models_as_dot_format=self.save_models_as_dot_format,
                                          root_dir=self.root_dir)
                self.created_models[name] = keras_model

    def create_models(self, inputs: tf.keras.Input) -> None:
        """
        Create different neural networks according to {self.models}

        Args:
            inputs: The input layer for the neural network.

        Returns:
            None
        """
        self.inputs = inputs
        # Create LSTM networks
        self.__create_lstm_networks()
        # Create GRU networks
        self.__create_gru_networks()
        # Create feed-forward neural networks
        self.__create_ffn()
        # Create CNN networks
        self.__create_cnn_networks()
        # Create bidirectional LSTM network
        self.__create_bi_lstm()
        # Create CNN-RNN hybrid network
        self.__create_cnn_rnn()
        # Create RNN Encoder-Decoder network
        self.__create_rnn_encoder_decoder()
        # Create Convolutional LSTM network
        self.__create_conv_lstm1d()
        # Create Encoder-Decoder network with a Self-Attention mechanism
        self.__create_encoder_decoder_self_attention()

    def __get_instances(self, model_type: str) -> int:
        # Loop through the model list to find the model of the given type
        for model in self.models:
            if model[0] == model_type:
                # Return the number of instances of the model
                return model[1]

        # Return 0 if no instance of the model type was found
        return 0



import multiprocessing as mp
from datetime import timedelta
from math import ceil
from typing import Tuple

import numpy as np
import pandas as pd
from scipy.signal import find_peaks

from eventdetector_ts import MIDDLE_EVENT_LABEL, TimeUnit, config_dict
from eventdetector_ts.data.helpers import get_timedelta, get_total_units
from eventdetector_ts.optimization import logger
from eventdetector_ts.optimization.algorithms import convolve_with_gaussian_kernel


class OptimizationData:
    """
    OptimizationData class represents the data used for the event extraction pipeline.

     Attributes:
        - time_unit (TimeUnit): Unit of time used in the dataset.
        - true_events (pd.DataFrame): DataFrame to store true events.
        - predicted_op (np.ndarray): Array to store predicted outcomes.
        - delta (int): The maximum time tolerance used to determine the correspondence between a predicted event
                    and its actual counterpart in the true events.
        - s_h (float): A step parameter for the peak height threshold h.
        - s_s (int): Step size in time unit for overalapping the partition.
        - w_s (int): Size in time unit of the overalapping partition.
        - t_max (float): The maximum total time related to sigma.
        - output_dir (str): The parent directory.
        - big_sigma (int): Value calculated based on t_max, w_s, and s_s.
        - overlapping_partitions (np.ndarray): Array to store overalapping partitions.

    """

    def __init__(self, t_max: float, w_s: int, s_s: int,
                 s_h: float,
                 delta: int,
                 output_dir: str, time_unit: TimeUnit):
        """
        Initializes the OptimizationData object.

        Args:
            t_max (float): The maximum total time related to sigma.
            w_s (int): Size in time unit of the overalapping partition.
            s_s (int): Step size in time unit for overalapping the partition.
            s_h (float): A step parameter for the peak height threshold h.
            delta (int): The maximum time tolerance used to determine the correspondence between a predicted event
                    and its actual counterpart in the true events.
            output_dir (str): The parent directory.
            time_unit (TimeUnit): Unit of time used in the dataset.
        """
        self.time_unit = time_unit
        self.true_events: pd.DataFrame = pd.DataFrame()
        self.predicted_op: np.ndarray = np.empty(shape=(0,))
        self.delta = delta
        self.s_h = s_h
        self.s_s = s_s
        self.w_s = w_s
        self.t_max = t_max
        self.output_dir = output_dir
        self.big_sigma = 1 + ceil((self.t_max - self.w_s) / self.s_s)
        self.overlapping_partitions: np.ndarray = np.empty(shape=(0,))

    def set_true_events(self, true_events: pd.DataFrame) -> None:
        self.true_events = true_events

    def set_overlapping_partitions(self, overlapping_partitions: np.ndarray):
        self.overlapping_partitions = overlapping_partitions

    def set_predicted_op(self, predicted_op: np.ndarray):
        self.predicted_op = predicted_op
        overlapping_partitions_test = self.overlapping_partitions[-len(predicted_op):]
        self.overlapping_partitions = overlapping_partitions_test
        first_partition_test_data = self.overlapping_partitions[0]
        last_partition_test_data = self.overlapping_partitions[-1]
        start_date_test_data = first_partition_test_data[0][-1].to_pydatetime()
        end_date_test_data = last_partition_test_data[0][-1].to_pydatetime()
        logger.info(
            f"Starting and ending dates of test data are respectively {start_date_test_data} --> {end_date_test_data}")

        true_events_test = self.true_events[(self.true_events[MIDDLE_EVENT_LABEL] >= start_date_test_data) & (
                self.true_events[MIDDLE_EVENT_LABEL] <= end_date_test_data)]
        self.true_events = true_events_test


def get_peaks(h: float, t: np.ndarray, op_g: np.ndarray) -> np.ndarray:
    """
    Compute peaks for given mid_times of partitions, op values and threshold h. 
    Args:
        h (float): Threshold for peaks.
        t (np.ndarray): mid_times of partitions
        op_g (np.ndarray): op values

    Returns:
        np.ndarray: Peaks.
    """
    peaks, _ = find_peaks(op_g, height=np.array([h, 1.0]))
    return t[peaks]


def compute_op_as_mid_times(overlapping_partitions: np.ndarray, op_g: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute op as a function of mid-times of partitions instead of partition's index.
    Args:
        overlapping_partitions (np.ndarray): overalapping partitions
        op_g (np.ndarray): Op array

    Returns:
        Tuple[np.ndarray, np.ndarray]: mid-times of partitions, op as a function of mid-times of partitions
    """
    t = []
    op_g_ = []
    for n in range(len(op_g)):
        w_n = overlapping_partitions[n]
        b_n = w_n[0][-1].to_pydatetime()
        e_n = w_n[-1][-1].to_pydatetime()
        c_n = b_n + (e_n - b_n) / 2
        t.append(c_n)
        op_g_.append(op_g[n])
    t, op_g_ = np.array(t), np.array(op_g_)
    return t, op_g_


class OptimizationCalculator:
    def __init__(self, optimization_data: OptimizationData):
        self.optimization_data = optimization_data

    def apply_gaussian_filter(self, sigma: int, m: int) -> np.ndarray:
        return convolve_with_gaussian_kernel(self.optimization_data.predicted_op, sigma, m=m)

    def __compute_op_as_mid_times(self, op_g: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        return compute_op_as_mid_times(self.optimization_data.overlapping_partitions, op_g)

    def __util_method(self, s_peaks: np.ndarray, delta_with_time_unit: timedelta) -> Tuple[int, int, int, list]:
        """
        Useful method for compute_f1score method.
        Args:
            s_peaks (np.ndarray): peaks of op.
            delta_with_time_unit (timedelta): delta as number in unit time.

        Returns:
            tp, fp, fn, delta_t
        """
        e_t = self.optimization_data.true_events.copy()

        fp: int = 0
        tp: int = 0
        delta_t: list = []
        for m_p in s_peaks:
            signed_delta = delta_with_time_unit
            t_t = None
            for i, t_e in enumerate(e_t[MIDDLE_EVENT_LABEL]):
                m_t = t_e
                diff = m_p - m_t

                if abs(diff) <= delta_with_time_unit:
                    if t_t is None or abs(m_p - t_t) > abs(diff):
                        t_t = m_t
                        signed_delta = diff

            if t_t is not None:
                tp += 1
                e_t = e_t.drop(e_t[e_t[MIDDLE_EVENT_LABEL] == t_t].index)
                diff = get_total_units(timedelta_=signed_delta, unit=self.optimization_data.time_unit)

                delta_t.append(diff)
            else:
                fp += 1
        fn: int = len(e_t)
        return tp, fp, fn, delta_t

    def compute_f1score(self, sigma: int, m: int, h: float):
        delta_with_time_unit = get_timedelta(self.optimization_data.delta, self.optimization_data.time_unit)
        op_g: np.ndarray = self.apply_gaussian_filter(sigma=sigma, m=m)
        t, op_g = self.__compute_op_as_mid_times(op_g=op_g)
        s_peaks = get_peaks(h=h, t=t, op_g=op_g)
        tp, fp, fn, delta_t = self.__util_method(s_peaks=s_peaks, delta_with_time_unit=delta_with_time_unit)

        if tp + fp == 0 or tp + fn == 0:
            return 0.0, 0.0, 0.0, [], []

        precision = tp / (tp + fp)
        recall = tp / (tp + fn)
        if precision + recall == 0:
            return 0.0, 0.0, 0.0, [], []
        return (2.0 * precision * recall) / (precision + recall), precision, recall, s_peaks.tolist(), delta_t

    def evaluate_combination(self, combination):
        sigma, m, h = combination
        f1_score, precision, recall, peaks, delta_t = self.compute_f1score(sigma, m, h)
        formatted_combination = ', '.join(f'{item:.2f}' for item in combination)
        logger.info(f"Evaluated Combination [sigma, m, h] : [{formatted_combination}] => [F1 Score: {f1_score:.4f}, "
                    f"Precision: {precision:.4f}, Recall: {recall:.4f}]")
        return f1_score, precision, recall, peaks, delta_t


class EventOptimization:
    """
    After obtaining the predicted op values from the metamodel, they are then processed
        through an optimization algorithm to extract the predicted events. This involves applying
        a Gaussian filter to smooth out the predictions and identifying peaks in the resulting signal
        that correspond to the mid-times of the predicted events, which are then compared to the
        actual events in the test set. The performance of the algorithm is evaluated by computing
        metrics such as F1-Score, which combines precision and recall using their harmonic means.
        Maximizing the F1-Score is the preferred metric for evaluating models since it requires
        simultaneously maximizing precision and recall.
    """

    def __init__(self, optimization_data: OptimizationData):
        self.optimization_data = optimization_data
        self.optimization_calculator: OptimizationCalculator = OptimizationCalculator(self.optimization_data)

    def max_f1score(self, use_multiprocessing: bool = False) -> tuple[list, list]:
        """
        The optimization process aims to maximize the F1-Score metric by fine-tuning several parameters,
            including the filter size (2m + 1) and standard deviation (σ) of the Gaussian filter,
            and the peak height threshold h.

        Args:
            use_multiprocessing (bool): Using or not multiprocessing

        Returns:
              list of peaks
        """
        sigma_range = range(1, self.optimization_data.big_sigma + 1)
        h_values = np.arange(0, 1, self.optimization_data.s_h)
        # Create a list of all combinations to evaluate
        combinations = [(sigma, m, h) for sigma in sigma_range for m in [sigma, 2 * sigma, 3 * sigma] for
                        h in h_values]

        try:
            if use_multiprocessing:
                # Create a multiprocessing pool with the desired number of processes
                num_processes = mp.cpu_count() // 2  # Use the number of available CPU cores
                pool = mp.Pool(processes=num_processes)
                # Evaluate combinations in parallel
                results = pool.map(self.optimization_calculator.evaluate_combination, combinations)
            else:
                # Evaluate combinations sequentially
                results = [self.optimization_calculator.evaluate_combination(combination) for combination in
                           combinations]
        except ValueError as e:
            logger.error(e)
            exit(0)

        # Find the combination with the maximum F1 score
        best_combination_index = np.argmax(list(map(lambda metrics: metrics[0], results)))
        best_combination = combinations[best_combination_index]
        config_dict["best_combination"] = best_combination
        max_f1_score, precision, recall, peaks, delta_t = results[best_combination_index]

        formatted_combination = ', '.join(f'{item:.2f}' for item in best_combination)
        logger.info(
            f"Best Combination [sigma, m, h] : [{formatted_combination}] => "
            f"[Max F1 Score: {max_f1_score:.4f} => Precision:{precision:.4f}, Recall:{recall:.4f}]")
        return peaks, delta_t



import csv
import os
from typing import Dict

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.patches import Patch

from eventdetector_ts import OUTPUT_DIR, TimeUnit, MIDDLE_EVENT_LABEL
from eventdetector_ts.data.helpers import get_timedelta
from eventdetector_ts.plotter import logger, COLOR_TRUE, COLOR_PREDICTED, STYLE_PREDICTED, STYLE_TRUE, FIG_SIZE, PALETTE
from eventdetector_ts.plotter.helpers import event_to_rectangle


class Plotter:
    """
    The Plotter class is responsible for generating and saving plots of the predicted and true op, events, delta_t,...
        It provides a convenient way to visualize and compare the performance of a
        predictive model against the actual observed values.
    """

    def __init__(self, root_dir: str, time_unit: TimeUnit, w_s: int) -> None:
        """
        Initialize the Plotter object.

        Args:
            root_dir (str): The root directory for saving the plots.
            time_unit (TimeUnit): The unit time of the dataset
            w_s (int): The width of each event in time unit
        """

        self.val_losses = {}
        self.train_losses = {}
        self.val_loss_meta_model: list = []
        self.train_loss_meta_model: list = []
        self.w_s = w_s
        self.time_unit = time_unit
        # Whether to display the plots or not. Defaults to False.
        self.show = True
        self.root_dir = root_dir
        self.predicted_y: np.ndarray = np.empty(shape=(0,))
        self.test_y: np.ndarray = np.empty(shape=(0,))
        self.predicted_events: list = []
        self.true_events: pd.DataFrame = pd.DataFrame()
        self.delta_t: list = []
        self.working_dir = os.path.join(root_dir, OUTPUT_DIR)
        os.makedirs(self.working_dir)

    def set_show(self, show: bool) -> None:
        """
        Set show value
        Args:
            show (bool): Value to set for 'self.show'

        Returns:
            None
        """
        self.show = show

    def set_data_op(self, test_y: np.ndarray, predicted_y: np.ndarray) -> None:
        """
        Set test_y and predicted_y
        Args:
            test_y: The true op values
            predicted_y: The predicted op values

        Returns:
            None
        """
        self.test_y = test_y
        self.predicted_y = predicted_y

    def set_data_events(self, predicted_events: list, true_events: pd.DataFrame) -> None:
        """
        Set true and predicted events
        Args:
            predicted_events (list): List of predicted events computed by the optimization process
            true_events (pd.DataFrame): DataFrame of true events

        Returns:
            None
        """
        self.predicted_events = predicted_events
        self.true_events = true_events

    def set_delta_t(self, delta_t: list) -> None:
        """
        Set delta_t
        Args:
            delta_t (list): Each item of this list contains the accepted delta in time unit between
                true event its correspondent in the list of predicted events

        Returns:
            None
        """
        self.delta_t = delta_t

    def set_losses(self, train_losses: Dict[str, list], val_losses: Dict[str, list],
                   train_loss_meta_model: list, val_loss_meta_model: list) -> None:
        """
        Set losses of all trained models.
        Args:
            train_losses (Dict[str, list]): train losses.
            val_losses (Dict[str, list]): val losses.
            train_loss_meta_model (list): train loss for the metamodel.
            val_loss_meta_model (list): val loss for the metamodel.
        Returns:
            None
        """
        self.train_losses = train_losses
        self.val_losses = val_losses
        self.train_loss_meta_model = train_loss_meta_model
        self.val_loss_meta_model = val_loss_meta_model

    def plot_prediction(self) -> None:
        """
        Plot the true and the predicted op and save it.

        Returns:
            None
        """

        logger.info("Plotting and saving the figure displaying the true and the predicted op")
        # Create the plot using Seaborn
        # Set the ggplot style
        sns.set(style="ticks", palette=PALETTE)
        plt.figure(figsize=FIG_SIZE)  # Set the figure size
        # Plot the true and predicted values using Seaborn
        n = len(self.test_y)
        sns.lineplot(x=np.arange(n), y=self.test_y, color=COLOR_TRUE, label='True Op')
        sns.lineplot(x=np.arange(n), y=self.predicted_y, color=COLOR_PREDICTED, label='Predicted Op')
        # Add labels and title to the plot
        plt.xlabel('Partitions')
        plt.ylabel('Op')
        plt.title('True Op vs Predicted Op')
        # Add legend
        plt.legend()
        # Save the plot to a file
        path = os.path.join(self.working_dir, "op.png")
        plt.savefig(path, dpi=300)
        # Show the plot
        if self.show:
            plt.show()
        self.__save_op()

    def plot_predicted_events(self) -> None:
        """
        Plot the true and the predicted events and save it.

        Returns:
            None
        """

        logger.info("Plotting and saving the figure displaying the true events and the predicted events")
        fig, ax = plt.subplots(figsize=FIG_SIZE)
        sns.set(style="ticks", palette=PALETTE)

        for i, predicted_event in enumerate(self.predicted_events):
            rect1 = event_to_rectangle(event=predicted_event, w_s=self.w_s, time_unit=self.time_unit,
                                       color=COLOR_PREDICTED,
                                       style=STYLE_PREDICTED)
            ax.add_patch(rect1)

        for _, test_date in self.true_events[MIDDLE_EVENT_LABEL].iteritems():
            rect1 = event_to_rectangle(event=test_date, w_s=self.w_s, time_unit=self.time_unit, color=COLOR_TRUE,
                                       style=STYLE_TRUE)
            ax.add_patch(rect1)

        locator = mdates.AutoDateLocator(minticks=3)
        formatter = mdates.AutoDateFormatter(locator)
        ax.xaxis.set_major_locator(locator)
        ax.xaxis.set_major_formatter(formatter)

        start_time = self.true_events[MIDDLE_EVENT_LABEL].iloc[0]
        end_time = self.true_events[MIDDLE_EVENT_LABEL].iloc[-1]
        ax.set_xlim([start_time, end_time])
        ax.set_ylim([0.0, 1.01])

        predicted_patch = Patch(color=COLOR_PREDICTED, label='Predicted Events')
        true_patch = Patch(color=COLOR_TRUE, label='True Events')
        ax.legend(handles=[predicted_patch, true_patch], edgecolor="black")

        # Save the plot to a file
        path = os.path.join(self.working_dir, "events.png")
        plt.savefig(path, dpi=300)
        # Show the plot
        if self.show:
            plt.show()
        self.__save_events()

    def plot_delta_t(self, bins=30) -> None:
        """
        Plots a histogram for delta t.

        Args:
            bins (int): The number of bins in the histogram. Default is 10.

        Returns:
              None
        """
        sns.set(style="ticks", palette=PALETTE)
        plt.figure(figsize=FIG_SIZE)

        sns.histplot(self.delta_t, bins=bins, binrange=(-self.w_s, self.w_s))

        plt.xlabel(f'delta ({self.time_unit})')
        plt.ylabel('Number of events')

        std = np.std(self.delta_t)
        mu = np.mean(self.delta_t)

        plt.title(f'Histogram std = {std:.2f}, mu = {mu:.2f}')
        # Save the plot to a file
        path = os.path.join(self.working_dir, "delta_t.png")
        plt.savefig(path, dpi=300)
        # Show the plot
        if self.show:
            plt.show()

    def plot_losses(self):
        """
        Plot losses for all trained models.
        Returns:
            None
        """
        meta_model_was_used: bool = len(self.val_loss_meta_model) > 0

        sns.set(style="ticks", palette=PALETTE)
        if meta_model_was_used:
            fig, (ax1, ax2) = plt.subplots(1, 2)
            fig.set_size_inches((11, 8.5), forward=False)
        else:
            fig, ax1 = plt.subplots(figsize=FIG_SIZE)
        y_label = 'Loss'
        x_label = 'Epochs'
        colors = sns.color_palette(PALETTE, len(self.val_losses))
        lifestyle_val = '--'
        lifestyle_train = '-'
        for i, (model_name, val_loss) in enumerate(self.val_losses.items()):
            epochs = range(1, len(val_loss) + 1)
            train_loss = self.train_losses[model_name]
            ax1.plot(epochs, train_loss, linestyle=lifestyle_train, color=colors[i],
                     label='Training Loss - {}'.format(model_name))
            ax1.plot(epochs, val_loss, linestyle=lifestyle_val, color=colors[i],
                     label='Validation Loss - {}'.format(model_name))
            ax1.set_ylabel(y_label)
            ax1.set_xlabel(x_label)
            ax1.legend()

        if len(self.val_loss_meta_model) > 0:
            epochs_meta = range(1, len(self.val_loss_meta_model) + 1)
            ax2.plot(epochs_meta, self.train_loss_meta_model, linestyle=lifestyle_train, color='b',
                     label='Training Loss - Meta Model')
            ax2.plot(epochs_meta, self.val_loss_meta_model, linestyle=lifestyle_val, color='g',
                     label='Validation Loss - Meta Model')
            ax2.set_ylabel(y_label)
            ax2.set_xlabel(x_label)
            ax2.legend()

        fig.suptitle('Training and Validation Losses')
        plt.tight_layout()
        # Save the plot to a file
        path = os.path.join(self.working_dir, "losses.png")
        plt.savefig(path, dpi=300)
        # Show the plot
        if self.show:
            plt.show()

    def __save_events(self) -> None:
        """
        Save predicted events/true events to csv files.

        Returns:
            None
        """
        path = os.path.join(self.working_dir, "predicted_events.csv")
        radius = get_timedelta(self.w_s // 2, self.time_unit)
        with open(path, 'w', encoding='UTF8', newline='') as f:
            writer = csv.writer(f, delimiter=' ')
            for i in range(len(self.predicted_events)):
                predicted_event = self.predicted_events[i]
                start_time = predicted_event - radius
                end_time = predicted_event + radius
                writer.writerow([start_time.isoformat(), end_time.isoformat()])

        path = os.path.join(self.working_dir, "true_events.csv")
        with open(path, 'w', encoding='UTF8', newline='') as f:
            writer = csv.writer(f, delimiter=' ')
            for _, test_date in enumerate(self.true_events[MIDDLE_EVENT_LABEL]):
                start_time = test_date - radius
                end_time = test_date + radius
                writer.writerow([start_time.isoformat(), end_time.isoformat()])

    def __save_op(self) -> None:
        """
        Save predicted/true Op into csv file.

        Returns:
            None
        """
        df = pd.DataFrame({'True-Op': self.test_y, 'Predicted-Op': self.predicted_y})
        path = os.path.join(self.working_dir, "op.csv")
        df.to_csv(path, index=True, sep=" ")



from datetime import datetime, timedelta


class Interval:
    """
    Represents a time interval between two datetime objects. This class is used to model an event or partition in
    time-series.
    """

    def __init__(self, start_time: datetime, end_time: datetime):
        """
        Constructs an interval for a given start and end time.

        Args:
            start_time (datetime): The starting time of the interval.
            end_time (datetime): The ending time of the interval.
        """
        self.start_time = start_time
        self.end_time = end_time
        self.duration = self.end_time - self.start_time

    def __str__(self) -> str:
        """
        Returns a string representation of the interval in the format "start_time ---> end_time".

        Returns:
            str: A string representation of the interval.
        """
        return "{} ---> {}".format(self.start_time, self.end_time)

    def overlap(self, other: 'Interval') -> timedelta:
        """
        Computes the overlapping time (ot) between this interval and another interval.

        Args:
            other (Interval): Another interval to compare with.

        Returns:
            timedelta: The overlapping time between this interval and the other interval as a timedelta object.
        """
        overlap_start_time = max(self.start_time, other.start_time)
        overlap_end_time = min(self.end_time, other.end_time)
        overlap_duration = max(timedelta(0), overlap_end_time - overlap_start_time)
        return overlap_duration

    def overlapping_parameter(self, other: 'Interval') -> float:
        """
        Computes the overlapping parameter between this interval and another interval.

        Args:
            other (Interval): Another interval to compare with.

        Returns:
            float: A floating number between 0.0 and 1.0 representing the degree of overlap between the two intervals.
        """
        if other is None:
            return 0.0
        overlap_duration = self.overlap(other)
        total_duration = self.duration + other.duration - overlap_duration
        return overlap_duration / total_duration



import os
from enum import Enum
from logging import config
from typing import Dict, Optional
from urllib.request import urlretrieve

import pandas as pd
from tqdm import tqdm

TIME_LABEL = "time"
MIDDLE_EVENT_LABEL = "event"

LSTM = "LSTM"
GRU = "GRU"
CNN = "CNN"
RNN_BIDIRECTIONAL = "RNN_BIDIRECTIONAL"
CONV_LSTM1D = "CONV_LSTM_1D"
RNN_ENCODER_DECODER = "RNN_ENCODER_DECODER"
CNN_RNN = "CNN_RNN"
SELF_ATTENTION = "SELF_ATTENTION"
FFN = "FFN"

FILL_NAN_ZEROS = 'zeros'
FILL_NAN_FFILL = 'ffill'
FILL_NAN_BFILL = 'bfill'
FILL_NAN_MEDIAN = 'median'

TYPE_TRAINING_AVERAGE = 'average'
TYPE_TRAINING_FFN = 'ffn'
META_MODEL_NETWORK = "meta_model_ffn"
META_MODEL_SCALER = "meta_model_scaler"

# Define constants for scaler types
MIN_MAX_SCALER = "MinMaxScaler"
STANDARD_SCALER = "StandardScaler"
ROBUST_SCALER = "RobustScaler"

SCALERS_DIR = "scalers"
MODELS_DIR = "models"
OUTPUT_DIR = "output"
CONFIG_FILE = ".config.json"
# Store some important values for prediction
config_dict: Dict = {}


class TimeUnit(Enum):
    """
    An enumeration of different time units.

    Attributes:
        SECOND: The time unit is in seconds.
        MILLISECOND: The time unit is in milliseconds.
        MICROSECOND: The time unit is in microseconds.
        MINUTE: The time unit is in minutes.
        HOUR: The time unit is in hours.
        DAY: The time unit is in days.
        YEAR: The time unit is in years.
    """
    SECOND = "second"
    MILLISECOND = "millisecond"
    MICROSECOND = "microsecond"
    MINUTE = "minute"
    HOUR = "hour"
    DAY = "day"
    YEAR = "year"

    @classmethod
    def _missing_(cls, value):
        return cls.SECOND

    def __str__(self):
        return self.value


LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "colored": {
            "()": "colorlog.ColoredFormatter",
            "format": "%(asctime)s %(log_color)s[%(levelname)s] %(name)s: %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
            "log_colors": {
                "DEBUG": "cyan",
                "INFO": "white",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red,bg_white",
            },
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "colored",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "": {
            "handlers": ["console"],
            "level": "INFO",
            "propagate": False,
        },
    },
}

config.dictConfig(LOGGING_CONFIG)


def my_hook(t):
    """
    Wraps tqdm instance. Don't forget to close() or __exit__()
    the tqdm instance once you're done with it (easiest using `with` syntax).
  
    Example
    -------
  
    
  
    """
    last_b = [0]

    def inner(b=1, bsize=1, t_size=None):
        """
        b  : int, optional
            Number of blocks just transferred [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        t_size  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if t_size is not None:
            t.total = t_size
        t.update((b - last_b[0]) * bsize)
        last_b[0] = b

    return inner


def load_dataset(file_path: str, name: str, url=None, index_col: Optional[int] = 0) -> pd.DataFrame:
    """
    Load a dataset from a file. If the file is not found, it will be downloaded from the given URL.

    Args:
        name: Name of the file to load
        index_col: the same value as pandas index_col
        file_path (str): The path to the dataset file.
        url (str): The URL from which to download the dataset (optional).

    Returns:
        pandas.DataFrame: The loaded dataset.
    """

    file_extension = os.path.splitext(file_path)[1].lower()

    if not os.path.isfile(file_path) and url:
        # Dataset file isn't found, download it
        with tqdm(unit='B', unit_scale=True, leave=True, miniters=1,
                  desc=f"Downloading {name}") as t:  # all optional kwargs
            urlretrieve(url, filename=file_path,
                        reporthook=my_hook(t), data=None)

    if file_extension == ".csv":
        # Read CSV file
        dataset = pd.read_csv(file_path, index_col=index_col)
    elif file_extension == ".pkl":
        # Read Pickle file
        dataset = pd.read_pickle(file_path)
    else:
        raise ValueError(f"Unsupported file format: {file_extension}")

    # Return the loaded dataset
    return dataset


def load_martian_bow_shock():
    """
        Load the Martian bow shock dataset and events, for more information check this link:  http://amda.cdpp.eu/

        Returns:
            A dataset and events as pd.DataFrame

        """
    url_dataset = "https://archive.org/download/martian_bow_shock_dataset/martian_bow_shock_dataset.pkl"
    url_events = "https://archive.org/download/martian_bow_shock_events/martian_bow_shock_events.csv"
    data_set = load_dataset(file_path="martian_bow_shock_dataset.pkl", name="Martian Bow Shock data set",
                            url=url_dataset)
    events = load_dataset(file_path="martian_bow_shock_events.csv", name="Martian Bow Shock events", index_col=None,
                          url=url_events)

    return data_set, events


def load_credit_card_fraud():
    """
    Load the credit card fraud dataset and events, for more information check this link: https://www.kaggle.com/datasets/mlg-ulb/creditcardfraud
    
    Returns:
        A dataset and events as pd.DataFrame

    """
    url_dataset = "https://archive.org/download/credit_card_fraud_dataset/credit_card_fraud_dataset.csv"
    url_events = "https://archive.org/download/credit_card_fraud_events/credit_card_fraud_events.csv"

    data_set = load_dataset(file_path="credit_card_fraud_dataset.csv", name="Credit Card Fraud data set",
                            url=url_dataset)
    events = load_dataset(file_path="credit_card_fraud_events.csv", name="Credit Card Fraud events", index_col=None,
                          url=url_events)

    return data_set, events
